# *********
# |docname|
# *********

from runestone import cmap
import sys
import os
import shutil
import getpass
import six
import click
import pathlib
import re
import subprocess
from paver.easy import sh
import importlib.metadata
import importlib.resources
import codecs
from runestone.server import get_dburl
from rsptx.build_tools.core import update_library, populate_static, manifest_data_to_db
from rsptx.cl_utils.core import load_project_dotenv


def lets_exit(code=0):
    click.echo(
        click.style(
            "The runestone command is deprecated.  You should convert your book to PreTeXt.",
            fg="yellow",
            bold=True,
        ),
        color=True,
    )
    sys.exit(code)


def deprecation_warning():
    click.echo(
        click.style(
            """The runestone command and RST extensions are deprecated.
You should convert your book to PreTeXt.
We will continue to add features to the interactive components, but
you will only have access to them through PreTeXt.""",
            fg="yellow",
            bold=True,
        ),
        color=True,
    )


if len(sys.argv) == 2:
    if "--version" in sys.argv:
        version = importlib.metadata.version("runestone")
        print("Runestone version {}".format(version))
        lets_exit()

load_project_dotenv()


@click.group(chain=True)
@click.option("--version", is_flag=True, help="More print version and exit")
def cli(version):
    """
    Usage: runestone [--version] subcommand

    """
    if version:
        version = importlib.metadata.version("runestone")
        print("Runestone version {}".format(version))
        lets_exit()


@cli.command()
def init():
    template_base_dir = pathlib.Path(
        importlib.resources.files("runestone.common").joinpath("project_template")
    ).absolute()
    config_stuff = importlib.resources.read_text(
        "runestone.common.project_template", "conf.tmpl"
    )
    paver_stuff = importlib.resources.read_text(
        "runestone.common.project_template", "pavement.tmpl"
    )
    conf_dict = {}
    print("This will create a new Runestone project in your current directory.")
    click.confirm("Do you want to proceed? ", abort=True, default=True)
    print(
        "Next we need to gather a few pieces of information to create your configuration files"
    )
    conf_dict["dynamic_pages"] = click.prompt(
        "Build book for dynamic page service?", type=bool, default=False
    )
    if conf_dict["dynamic_pages"] == False:
        conf_dict["use_services"] = click.prompt(
            "Use Runestone Web Services ", default="false"
        )
    else:
        conf_dict["use_services"] = "true"
    conf_dict["author"] = click.prompt("Your Name ", default=getpass.getuser())
    conf_dict["project_title"] = click.prompt(
        "Title for this project ", default="Runestone Default"
    )
    conf_dict["python3"] = click.prompt(
        "Use Simple Python3 Semantics ", default="false"
    )
    conf_dict["default_ac_lang"] = click.prompt(
        "Default ActiveCode language", default="python"
    )
    if conf_dict["use_services"] == "true":
        conf_dict["project_name"] = (
            "os.path.basename(os.path.dirname(os.path.abspath(__file__)))"
        )
        conf_dict["build_dir"] = "./build"
        conf_dict["dest"] = "./published"
        conf_dict["login_req"] = click.prompt("Require login ", default="false")
        conf_dict["master_url"] = ""
        conf_dict["log_level"] = (
            10 if click.prompt("Log student actions? ", type=bool, default=True) else 0
        )
        conf_dict["dburl"] = click.prompt(
            "DataBase Connection URL",
            default="postgresql://user:password@localhost/runestone",
        )
        conf_dict["enable_chatcodes"] = click.prompt(
            "Enable Enable the chatcode feature)", type=bool, default=False
        )
        # See the comments in ``conf.tmpl`` on server-side grading for an explanation of these conditions.
        if conf_dict["log_level"] and conf_dict["login_req"]:
            conf_dict["server_side_grading"] = click.prompt(
                "Grade questions on the server where possible?",
                type=bool,
                default=False,
            )
        else:
            conf_dict["server_side_grading"] = False
        conf_dict["allow_pairs"] = click.prompt(
            "Enable Pair Programming feature(s)", type=bool, default=False
        )
    else:
        conf_dict["project_name"] = click.prompt("Project name: (one word, no spaces)")
        while " " in conf_dict["project_name"]:
            conf_dict["project_name"] = click.prompt(
                "Project name: (one word, NO SPACES)"
            )
        # Add quotes around the project name for use in the template.
        conf_dict["project_name"] = repr(conf_dict["project_name"])
        conf_dict["build_dir"] = click.prompt("Path to build dir ", default="./build")
        conf_dict["dest"] = click.prompt(
            "Path to deploy built site ", default="../../static"
        )
        conf_dict["login_req"] = "false"
        conf_dict["master_url"] = "http://127.0.0.1:8000"
        conf_dict["log_level"] = 0
        conf_dict["dburl"] = ""
        conf_dict["enable_chatcodes"] = "false"
        conf_dict["server_side_grading"] = False
        conf_dict["allow_pairs"] = "false"
    conf_dict["short_name"] = conf_dict["project_name"]
    conf_dict["downloads_enabled"] = click.prompt(
        "Enable inline Activecode downloads by default (single activecode downloads may be enabled with the :enabledownload: flag)",
        default="false",
    )

    shutil.copytree(os.path.join(template_base_dir, "_sources"), "_sources")
    os.makedirs(conf_dict["build_dir"])
    paver_final = paver_stuff.decode("utf-8") % conf_dict
    config_final = config_stuff.decode("utf-8") % conf_dict

    # On Windows, Python 3.6, the bytes read from ``template_base_dir`` and ``config_stuff`` contain Windows-style ``\n\r``. Unfortunately, `resource_string <http://setuptools.readthedocs.io/en/latest/pkg_resources.html#basic-resource-access>`_ does no universal newline support, so these remain intact. When written out, this is changed to ``\n\n``, making the file double-spaced. Python 3's `StringIO <https://docs.python.org/3/library/io.html#io.StringIO>`_ class provides universal newline support, while Python 2's `StringIO <https://docs.python.org/2/library/stringio.html#StringIO.StringIO>`__ doesn't.
    if six.PY3:
        # Per the `TextIOWrapper docs <https://docs.python.org/3/library/io.html#io.TextIOWrapper>`_, ``newline=None`` selects universal newline mode. The Python 3 StringIO_ class's ``newline`` argument works the same.
        paver_final = six.StringIO(paver_final, newline=None).read()
        config_final = six.StringIO(config_final, newline=None).read()

    with codecs.open("pavement.py", "w", encoding="utf8") as pvf:
        pvf.write(paver_final)

    with codecs.open("conf.py", "w", encoding="utf8") as pvf:
        pvf.write(config_final)

    print("Done. Type runestone build to build your project")


@cli.command()
@click.option("--all/--not-all", default=False, help="build all")
@click.option("--wd", default=None, help="change working directory before build")
def build(all, wd):
    from paver.tasks import main as paver_main

    if wd:
        os.chdir(wd)
    else:
        os.chdir(findProjectRoot())
    sys.path.insert(0, os.getcwd())
    try:
        with importlib.resources.path(
            "runestone.dist", "webpack_static_imports.json"
        ) as p:
            if not pathlib.Path(p).exists():
                click.echo(
                    "Error -- you are missing webpack_static_imports.json.  Please make sure"
                )
                click.echo(
                    "you have run projects/interactives/build.py to generate this file"
                )
                lets_exit(-1)
    except ModuleNotFoundError:
        click.echo(
            "Error -- you are missing webpack_static_imports.json.  Please make sure"
        )
        click.echo("you have runestone and the node dependencies installed correctly.")
        click.echo(
            "In a development environment, execute npm run build in bases/rsptx/interactives"
        )
        lets_exit(-1)

    version = importlib.metadata.version("runestone")
    print("Building with Runestone {}".format(version))

    import pavement

    confpath = pathlib.Path(pavement.options.get("confdir", "."))

    if pavement.options.build.template_args["basecourse"] != pavement.project_name:
        click.echo(
            click.style(
                f"""Please update pavement.py so that basecourse ({pavement.options.build.template_args["basecourse"]})
                            and project_name ({pavement.project_name}) match""",
                fg="red",
            ),
            err=True,
            color=True,
        )
        lets_exit(1)
    with open(confpath / "conf.py", encoding="utf-8") as cf:
        ctext = cf.read()
        if not re.search(r"from runestone import.*(setup|script_files)", ctext):
            click.echo(
                click.style(
                    """Please update conf.py to import setup or script_files and css_files
                    from runestone. If you do not import setup you must provide your own.""",
                    fg="red",
                ),
                err=True,
                color=True,
            )
            click.echo(
                "The current line probably looks like:\nfrom runestone import runestone_static_dirs, runestone_extensions"
            )
            click.echo(
                "Change it to:\nfrom runestone import runestone_static_dirs, runestone_extensions, setup"
            )
            lets_exit(1)

    if not os.path.exists(pavement.options.build.builddir):
        os.makedirs(pavement.options.build.builddir)

    myargs = ["build"]
    if all:
        myargs.append("--all")

    paver_main(args=myargs)

    if pavement.options.build.template_args.get("dynamic_pages", False):
        # Need a small config object with a dburl attribute just for update_library
        config = type("config", (object,), {})()
        config.dburl = get_dburl()
        course = ""
        if hasattr(pavement.options, "project_name"):
            course = pavement.options.project_name
        if not course:
            course = pavement.options.template_args["course_id"]
        update_library(config, "", course, click, build_system="Runestone")

    deprecation_warning()


@cli.command(short_help="preview the book in a minimal server (NO API support)")
@click.option("--port", default=8000, help="port for server to listen on")
@click.option("--listen", default="", help="address for server to listen on")
def preview(port, listen):
    _preview(port, listen)


def _preview(port, listen):
    click.echo("Note: this is a minimal static server without templates or a database.")
    click.echo("For many use cases this is fine.")
    click.echo(
        "For the full server, see https://github.com/RunestoneInteractive/RunestoneServer"
    )
    os.chdir(findProjectRoot())
    sys.path.insert(0, os.getcwd())
    try:
        import pavement

        try:
            if pavement.dynamic_pages == True:
                click.echo(
                    click.style(
                        """Error -- dynamic_pages is True, but this preview server does not support templates.
                    Please edit pavement.py and set dynamic_pages=False""",
                        color="red",
                    ),
                    err=True,
                )
                click.echo("You should update pavement.py and rebuild")
                return
        except:
            click.echo("dynamic_pages is not defined")
    except:
        print("Error, you must be in your project root directory")
        return

    os.chdir(pavement.serving_dir)

    if six.PY2:
        import SimpleHTTPServer
        import SocketServer

        Handler = SimpleHTTPServer.SimpleHTTPRequestHandler
        httpd = SocketServer.TCPServer((listen, port), Handler, bind_and_activate=False)
    else:
        import http.server
        import socketserver

        Handler = http.server.SimpleHTTPRequestHandler
        httpd = socketserver.TCPServer((listen, port), Handler, bind_and_activate=False)

    print("serving at port", port)
    httpd.allow_reuse_address = True
    httpd.server_bind()
    httpd.server_activate()
    sys.stderr = open("runestone.log", "a")
    httpd.serve_forever()


# configure preview as an alias for serve
@cli.command(short_help="Deprecated - use preview")
@click.option("--port", default=8000, help="port for server to listen on")
@click.option("--listen", default="", help="address for server to listen on")
def serve(port, listen):
    click.echo("The serve command is deprecated, use runestone preview")
    _preview(port, listen)


@cli.command()
@click.option("--dest", default="", help="destination for deploy")
def deploy(dest):
    os.chdir(findProjectRoot())
    sys.path.insert(0, os.getcwd())
    try:
        import pavement
    except ImportError as ie:
        print("Error, you must be in your project root directory")

    if not dest:
        try:
            dest = pavement.dest
        except:
            raise IOError(
                "No destination configured add dest to your pavement.py or use --dest"
            )

    click.echo("Deploying from " + pavement.serving_dir + " to " + dest)
    if os.name == "nt":
        # From ``robocopy /?``:
        #
        # /MIR  MIRror a directory tree (equivalent to /E plus /PURGE).
        #
        # /MT   Do multi-threaded copies with n threads (default 8).
        #
        # /NFL  No File List - don't log file names.
        #
        # Robocopy copies the contents of the source directory, not the source directory itself. So, append the final path of the source directory to the destination directory.
        sh(
            "robocopy /mir /mt /nfl {} {}".format(
                pavement.serving_dir,
                os.path.join(dest, os.path.basename(pavement.serving_dir)),
            ),
            ignore_error=True,
        )
    else:
        sh("rsync -rav --delete {} {}".format(pavement.serving_dir, dest))


@cli.command(short_help="Run sphinx build to convert RST to PreTeXt")
@click.option("--course", default=None, help="Unique name of the book")
@click.option("--sourcedir", default="_sources", help="Where is the source Luke?")
@click.option("--outdir", default="build", help="Where is the source Luke?")
def rs2ptx(course, sourcedir, outdir):
    """
    Assemble and run a sphinx command similar to the following:

    .. code-block:: bash

        sphinx-build -b xml -d ./build/overview/doctrees -c . -Acourse_id=overview -Alogin_required=false -Aappname=runestone -Aloglevel=10 -Acourse_url=https://runestone.academy -Adynamic_pages=True -Ause_services=true -Abasecourse=overview -Apython3=true -Adownloads_enabled=true -Aallow_pairs=false -Aenable_chatcodes=false -Arunestone_version=5.7.1 -Abuild_info=unknown . ./build/xml

    This command demonstrates a step toward independence from paver and the pavement.py file.
    This is kind of moot in a future where we rely on pretext to be the authoring language. But it **would**
    be easy to move the key variables and template_args from pavement.py to conf.py  and update the build
    command to work like this...
    """
    os.chdir(findProjectRoot())
    sys.path.insert(0, os.getcwd())
    if "DBURL" in os.environ:
        del os.environ["DBURL"]
    try:
        import pavement
    except Exception as e:
        click.echo("Could not read pavement.py file, aborting")
        click.echo(f"Details: {e}")
        lets_exit(1)

    if not course:
        if pavement.project_name:
            course = pavement.project_name
        else:
            course = click.prompt("Name of Course ", default="rsbook")

    if pavement.template_args:
        tdict = pavement.template_args
    else:
        tdict = {"course_id": course}

    cmd_start = [
        "sphinx-build",
        "-E",
        "-b",
        "xml",
        "-d",
        f"./build/{course}/doctrees",
        "-c",
        ".",
    ]
    tplate_val = [f"-A{key}={val}" for key, val in tdict.items()]
    cmd_end = [f"{sourcedir}", f"./{outdir}/xml"]
    cmd = " ".join(cmd_start + tplate_val + cmd_end)
    click.echo(cmd)
    try:
        cp = subprocess.run(cmd, shell=True, check=True)
    except subprocess.CalledProcessError as e:
        click.echo(f"{e.stderr or ''}{e.stdout or ''}")
        raise

    # TODO: Run the xsltproc command after a successful runestone convert


@cli.command(short_help="type runestone doc directive to get help on directive")
@click.option("--list", is_flag=True, help="List all commands")
@click.argument("directive", nargs=-1)
def doc(directive=None, list=None):
    """
    Show Format and all options for a runestone directive
    """
    if list:
        print("Runestone Directives List")
        print("  ", "\n   ".join(sorted(cmap.keys())))
        return

    if directive:
        directive = directive[0]
        if directive in cmap:
            print(cmap[directive].__doc__)
        else:
            print("""Unknown Directive.  Possible values are""")
            print("  ", "\n   ".join(sorted(cmap.keys())))


@cli.command(short_help="Update template files")
def update():
    """
    Update all template files

    Warning!  This is a destructive command.  If you have made local changes to your
    templates (especially and most probably layout.html) They will be moved aside
    you can merge your changes (in _templates.bak) after you run this command.
    """
    os.chdir(findProjectRoot())
    template_base_dir = pathlib.Path(
        importlib.resources.files("runestone.common").joinpath("project_template")
    ).absolute()
    print("This will update all files in the _templates folder.")
    print(
        "The old _templates folder will be in _templates.bak so you can merge manually after you update"
    )
    click.confirm("Do you want to proceed? ", abort=True, default=True)
    shutil.rmtree("_templates.bak", ignore_errors=True)
    shutil.move("_templates", "_templates.bak")
    shutil.copytree(os.path.join(template_base_dir, "_templates"), "_templates")


@cli.command(short_help="Process runestone-manifest.xml file")
@click.option("--course", help="Name of the course (base course)")
@click.option(
    "--manifest",
    default="runestone-manifest.xml",
    help="path to runestone-manifest.xml file",
)
def process_manifest(course, manifest):
    """Populate a runestone database with meta data about a course created with the PreTeXt processor

    Arguments:
        course {string} -- Name of the base course
        manifest {path} -- Path to manifest file

    Raises:
        IOError: If manifest file not found
    """
    if os.path.exists(manifest):
        manifest_data_to_db(course, manifest)
    else:
        raise IOError("You must provide a valid path to a manifest file")


@cli.command(short_help="Fetch Javascript/CSS from CDN and copy to _static")
@click.option("--course", help="Name of the course (base course)")
def fetch_latest_static(course):
    config = type("config", (object,), {})()
    config.dburl = get_dburl()
    os.chdir(findProjectRoot())
    manifest = "runestone-manifest.xml"
    mpath = pathlib.Path(os.getcwd(), "published", course, manifest)
    populate_static(config, mpath, course)


def main(args=None):
    sys.dont_write_bytecode = True
    if not args:
        args = sys.argv[1:]
    if not args:
        print("""Usage: runestone help for a list of commands""")
        lets_exit(0)
    cli.add_command(init)
    cli.add_command(build)
    cli.add_command(serve)
    cli.add_command(deploy)
    cli.add_command(doc)
    cli.add_command(update)
    cli()


def findProjectRoot():
    start = os.getcwd()
    prevdir = ""
    while start != prevdir:
        if os.path.exists(os.path.join(start, "pavement.py")):
            return start
        if os.path.exists(os.path.join(start, "project.ptx")):
            return start
        prevdir = start
        start = os.path.dirname(start)
    raise IOError("You must be in a runestone project to run runestone")


if __name__ == "__main__":
    lets_exit(main(sys.argv[1:]))



# *************************************************
# |docname| - Support wavedrom diagrams from Sphinx
# *************************************************
# See https://wavedrom.com/ for full info. Briefly, this library transforms timing diagrams written in a JSON format into its graphical equivalent.
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
from typing import List


# Third-party imports
# -------------------
from docutils.parsers.rst import directives
from docutils import nodes
from docutils.nodes import Node
from sphinx.ext.graphviz import figure_wrapper, align_spec
from sphinx.util.docutils import SphinxDirective
from sphinx.writers.html import HTMLTranslator

# Local imports
# -------------
# None.


# WaveDrom
# ========
# This code is based on parts of ``sphinx.ext.graphviz``. It basically stores WaveJSON in the appropriate HTML tag, then relies on the JS to turn this into a timing diagram.
class wavedrom(nodes.General, nodes.Inline, nodes.Element):
    pass


class WaveDromDirective(SphinxDirective):
    """
    Directive to insert arbitrary dot markup.
    """

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {
        "alt": directives.unchanged,
        "align": align_spec,
        "caption": directives.unchanged,
        "name": directives.unchanged,
        "class": directives.class_option,
    }

    def run(self) -> List[Node]:
        node = wavedrom()
        node["code"] = "\n".join(self.content)
        node["options"] = {"docname": self.env.docname}
        if "alt" in self.options:
            node["alt"] = self.options["alt"]
        if "align" in self.options:
            node["align"] = self.options["align"]
        if "class" in self.options:
            node["classes"] = self.options["class"]

        if "caption" not in self.options:
            self.add_name(node)
            return [node]
        else:
            figure = figure_wrapper(self, node, self.options["caption"])
            self.add_name(figure)
            return [figure]


# The core: store the waveJSON in the HTML then ask for a render.
def html_visit_wavedrom(self: HTMLTranslator, node: wavedrom) -> None:
    self.body.append(
        # Wrap this in a ``data-component`` tag so that wavedrom code will be dynamically loaded. See ``webpack.index.js``, which references ``js/wavedrom.js``.
        '<div data-component="wavedrom">'
        # Finally, insert the WaveJSON file for this diagram.
        '<script type="WaveDrom">{'
    )
    self.body.append(node["code"])
    self.body.append("}</script></div>")
    # Skip the depart visitor -- the HTML above is all we need.
    raise nodes.SkipNode


def setup(app):
    app.add_directive("wavedrom", WaveDromDirective)
    app.add_node(wavedrom, html=(html_visit_wavedrom, None))



from enum import Enum
from rsptx.data_types.question_type import QuestionType

class AutogradeOptions(Enum):
    MANUAL = ("manual", "Manual", QuestionType.all_types())
    ALL_OR_NOTHING = ("all_or_nothing", "All or nothing", QuestionType.all_types())
    PCT_CORRECT = ("pct_correct", "Pct correct", QuestionType.all_types())
    INTERACT = ("interact", "Interact", QuestionType.all_types())
    PEER = ("peer", "Peer", [QuestionType.MCHOICE])
    PEER_CHAT = ("peer_chat", "Peer chat", [QuestionType.MCHOICE])

    def to_dict(self):
        return {
            "value": self.value[0],
            "label": self.value[1],
            "supported_question_types": [qt.value_only() for qt in self.value[2]],
        }



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# This file runs during runestone build. It generates the node data
# that is used to render the component by the  quizly.js script.

# Note: An import entry for quizly must be included in runestone/__init__.py

# Note: The content files for the quizly component must stored in the
# course's _static folder. Download the following file and unzip it in _static:
#  https://github.com/ram8647/quizly/blob/35ee7c945e6f240450f46f41ad9c80735215b5e0/quizly-runestone.zip
#

import shutil
import os
from pathlib import Path
from runestone.server.componentdb import addQuestionToDB, addHTMLToDB
from runestone.common import RunestoneIdDirective, RunestoneIdNode
from docutils import nodes

__author__ = "rmorelli"

# Debug flags
DEBUG = False
VERBOSE = False


# Template that will load index.html?quizname=the-quiz-name into an <iframe>
# NOTE: Hardcoding the container class.  Temporarily??
QUIZLY_TEMPLATE = """
       <div class="runestone alert alert-warning %(optclass)s">
       <div data-component="quizly" id="%(divid)s" data-question_label="%(question_label)s" style="visibility: visible;">
         <iframe height="595" src="../_static/quizly/index.html?backpack=hidden&amp;selector=hidden&amp;quizname=###&amp;hints=true&amp;repeatable=false" style="border: 0px; margin: 1px; padding: 1px;" width="100%%">
         </iframe>
       </div>
       </div>
       """

# Define the quizly directive


def setup(app):
    app.add_directive("quizly", Quizly)
    app.add_node(QuizlyNode, html=(visit_quizly_html, depart_quizly_html))


# The only content needed from quizly.py is the quizname


class QuizlyNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.


def visit_quizly_html(self, node):

    node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])
    self.body.append(node["delimiter"])

    print("DEBUG: visit_quizly_html quizname = " + node["quizname"]) if DEBUG else None
    print("DEBUG: visit_quizly_html template = " + node["template"]) if DEBUG else None
    print(
        "DEBUG: visit_quizly_html options = " + str(node["runestone_options"])
    ) if DEBUG else None

    res = node["template"] % (node["runestone_options"])
    print("DEBUG: visit_quizly_html res = " + res) if DEBUG else None
    self.body.append(res)


def depart_quizly_html(self, node):
    """This is called at the start of processing an activecode node.  If activecode had recursive nodes
    etc and did not want to do all of the processing in visit_ac_html any finishing touches could be
    added here.
    """
    print("DEBUG: depart_quizly_html") if DEBUG else None
    bc = node["runestone_options"]["basecourse"]
    addHTMLToDB(
        node["runestone_options"]["divid"],
        bc,
        "".join(self.body[self.body.index(node["delimiter"]) + 1 :]).replace(
            "../_static", f"/ns/books/published/{bc}/_static"
        ),
    )
    self.body.remove(node["delimiter"])


def process_activcode_nodes(app, env, docname):
    pass


def purge_activecodes(app, env, docname):
    pass


# The object created when the quizly directive is parsed in an RST.
# The quizname is the only required argument.
class Quizly(RunestoneIdDirective):
    """
    .. quizly:: some_unique_id, e.g., q1

       :quizname: quiz_eval_expression
    """

    required_arguments = 1
    optional_arguments = 0
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update({})

    def run(self):
        super(Quizly, self).run()
        print("DEBUG; Quizly.run()") if DEBUG else None

        addQuestionToDB(self)  # Not sure whether this works?
        document = self.state.document
        rel_filename, filename = document.settings.env.relfn2path(self.arguments[0])

        pathDepth = rel_filename.count("/")
        self.options["quizlyHomePrefix"] = "./" * pathDepth

        plstart = len(self.content)
        if self.content:
            self.options["controls"] = self.content[:plstart]

        quizly_node = QuizlyNode()
        quizly_node["runestone_options"] = self.options
        quizly_node["quizname"] = str(self.options["controls"][0])
        quizly_node["quizname"] = str.strip(quizly_node["quizname"][10:])
        quizly_node["template"] = QUIZLY_TEMPLATE.replace(
            "###", quizly_node["quizname"]
        )

        (
            quizly_node["source"],
            quizly_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)
        print("DEBUG: run() self.content = " + str(self.content)) if DEBUG else None
        print("DEBUG: run() quizly_node = " + str(quizly_node)) if DEBUG else None
        return [quizly_node]



# ********************************************************
# |docname| - Add question numbers to Runestone components
# ********************************************************
# Question numbers are prefixed with the current section of the document. These section numbers aren't computed until the doctree is resolved, so the question numbers can't be determined until then.
#
# Import
# ======
# Third-party
from docutils import nodes
import pdb
# Local
from .runestonedirective import RunestoneIdNode
from ..server.componentdb import addQNumberToDB


# Code
# ====
# After the section numbers are available from the TOC tree, determine the section number for each Runestone component with an ID.
def _insert_qnum(app, doctree, docname):
    toc = app.env.toc_secnumbers.get(docname, {})
    # Return the section number tuple for the given ``section_ref`` if it exists, or None if not.

    def get_secnum_tuple(section_ref):
        # The Sphinx TOC structure is::
        #
        #   Dict[docname: str,
        #       Dict[ref: str, secnum_tuple: Tuple[int, ...]
        #   ]
        #
        # where a ``ref`` of ``''`` provides a tuple for the document as a whole.

        return toc.get(section_ref)

    # Set the current section number to the section for the overall document, if it's available; otherwise, use no number (an empty tuple).
    current_secnum_tuple = get_secnum_tuple("") or tuple()

    current_question_number = 1
    for node in doctree.traverse():
        if isinstance(node, nodes.section):
            # To create a reference for a section node, I used this string, which was taken from ``sphinx.environment.collectors.toctree.TocTreeCollector.process_doc.build_toc``.
            new_secnum_tuple = get_secnum_tuple('#' + node['ids'][0])
            # Question numbering restarts at the beginning of each numbered section; otherwise, continue to existing numbering through unnumbered sections.
            if new_secnum_tuple:
                current_secnum_tuple = new_secnum_tuple
                current_question_number = 1
        elif isinstance(node, RunestoneIdNode):
            question_number_tuple = current_secnum_tuple + (current_question_number, )
            question_number_str = ".".join(map(str, question_number_tuple))
            # Update the database.
            addQNumberToDB(app, node, question_number_str)
            try:
                div_id = node["runestone_options"]["divid"]
                node["runestone_options"]["question_label"] = question_number_str
            except:
                div_id = node.runestone_options["divid"]
                node.runestone_options["question_label"] = question_number_str
            # Prepare to number the next question.
            current_question_number += 1
        else:
            pass



# *********
# |docname|
# *********
# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "bmiller"

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst import Directive
import json
import random

#    setup is called in activecode.py


def textfield_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Usage:
    In your document you can write :textfield:`myid:myvalue:width`
    This will translate to:
        <input type='text' id='myid' class="form-control input-small" style="display:inline; width:width;" value='myvalue'></input>

    where width can be specified in pixels or percentage of page width (standard CSS syntax).
    Width can also be specified using relative sizes:
        mini, small, medium, large, xlarge, and xxlarge
    """
    iid, value, width = text.split(":")

    if "mini" in width:
        width = "60px"
    elif "small" in width:
        width = "90px"
    elif "medium" in width:
        width = "150px"
    elif "large" in width:
        width = "210px"
    elif "xlarge" in width:
        width = "270px"
    elif "xxlarge" in width:
        width = "530px"

    res = (
        """<input type='text' id='%s' class="form-control" style="display:inline; width: %s;" value="%s"></input>"""
        % (iid, width, value)
    )

    return [nodes.raw("", res, format="html")], []



# **********************************
# |docname| - Database configuration
# **********************************
# Set up database configuration in this file
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8`_.
#
# Standard library
# ----------------
# None.
#
# Third-party imports
# -------------------
# Use asyncio for SQLAlchemy -- see `SQLAlchemy Asynchronous I/O (asyncio) <https://docs.sqlalchemy.org/en/14/orm/extensions/asyncio.html>`_.
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy.sql import select

# Local application imports
# -------------------------
from rsptx.configuration.core import settings, BookServerConfig, DatabaseType
from rsptx.logging import rslogger


if settings.database_type == DatabaseType.SQLite:
    connect_args = {"check_same_thread": False}
else:
    connect_args = {}

# The polling in `../../test/test_runestone_components.py` produces a HUGE amount of output when echo is true.
extra_settings = (
    {}
    if settings.book_server_config == BookServerConfig.test
    else dict(echo=settings.db_echo)
)
engine = create_async_engine(
    settings.database_url, connect_args=connect_args, **extra_settings
)
# This creates the SessionLocal class.  An actual session is an instance of this class.
async_session = sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)

# This creates the base class we will use to create models
Base = declarative_base()


async def init_models():
    async with engine.begin() as conn:
        # Never, ever drop tables in a production environment!
        if (
            settings.book_server_config
            in [BookServerConfig.development, BookServerConfig.test]
            and settings.drop_tables == "Yes"
        ):
            await conn.run_sync(Base.metadata.drop_all)
        await conn.run_sync(Base.metadata.create_all)


# Look for any records that violate non-null constraints. TODO: when/where should we call this? Or should it be removed?
async def check_not_null():
    rslogger.info("Searching for NOT NULL constraint violations..."),
    not_null_count = 0
    async with async_session() as session:
        for table_name, table in Base.metadata.tables.items():
            for column in table.columns:
                if not column.nullable:
                    # SQLAlchemy requires ``==`` to correctly create the query; it can't overload the ``is`` operator.
                    query = select(table).where(column == None)  # noqa: E711.
                    res = (await session.execute(query)).fetchall()
                    if res:
                        not_null_count += 1
                        rslogger.error(
                            f"Column {table_name}.{column.key} has {len(res)} NULL records, such as:"
                        )
                        for row in res[0:9]:

                            def shorten(s):
                                s = str(s)
                                return s if len(s) < 20 else f"{s[0:20]}..."

                            # The result isn't an ORM object, so use this to display it.
                            s = ", ".join(f"{k}={shorten(row[k])}" for k in row.keys())
                            rslogger.error(f"  {s}")
    rslogger.info(f"Done; found {not_null_count} columns with constraint violations.")


# If the engine isn't disposed of, then a PostgreSQL database will remain in a pseudo-locked state, refusing to drop or truncate tables (see `bookserver_session`).
async def term_models():
    await engine.dispose()


# Dependency
async def get_session() -> AsyncSession:
    async with async_session() as session:
        yield session



# **************************************************************
# |docname| - pytest fixtures used by both the client and server
# **************************************************************
# The goal of this test framework is to enable tests written for the Runestone Components to be easily run on the server, where values stored by the database can also be checked. This avoids a duplication of test code between Runestone Components and the server.
#
# To accomplish this goal, this file provides functions that work on both client and server. Then, the client and server must each uniquely define two fixtures (see `conftest.py`) specalized for that environment which provided the same results (access to the Selenium WebDriver and to a class of Selenium utilities). Both client and server tests must import from this file; tests can then
#
# Structure:
#
# - The ``selenium_driver_session`` fixture performs client- and server-specific setup, then returns an instance of a Selenium WebDriver. Since it's specific to the environment, this fixture must be defined differently for client and server. It should be defined at the module level or higher, since starting Selenium is an expensive operation.
# - The shared `selenium_driver`_ fixture then sets up/tears down the Selenium environment for each test.
# - The client/server-specific ``selenium_utils`` fixture wraps the shared `_SeleniumUtils`_ class, which provides a common set of methods for the tests.
# - All client tests should use the ``selenium_utils`` fixture, ensuring that they can run in either environment.
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
import logging

# Third-party imports
# -------------------
import pytest
from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait

# Local imports
# -------------
# None.


# Code
# ====
#
# selenium_driver
# ---------------
# Provide a way to reset the state of Selenium for each test, without exiting/restarting the driver (which is slow).
@pytest.fixture()
def selenium_driver(selenium_driver_session):
    driver = selenium_driver_session
    # Copied from the Runestone Components test framework.
    driver.implicitly_wait(10)

    try:
        yield driver
    finally:
        # Print the logs -- see the setup in `selenium_logging <selenium_logging>`. Capture them if setup fails, since the logs may help understand the failure. However, pytest ignore any generated logs; see `my bug report <https://github.com/pytest-dev/pytest/issues/9021>`_. As a workaround, uncomment the print statement below to dump unfiltered logs during a test run.
        #
        # Transform Chrome log levels to `Python log levels <https://docs.python.org/3/library/logging.html#logging-levels>`_.
        chrome_to_py_loglevels = {
            "NOTSET": 0,
            "DEBUG": 10,
            "INFO": 20,
            "WARNING": 30,
            "ERROR": 40,
            "SEVERE": 40,
            "CRITICAL": 50,
        }
        py_logger = logging.getLogger("Chrome.JavaScript.console")
        chrome_logs = driver.get_log("browser")
        for log in chrome_logs:
            py_logger.log(chrome_to_py_loglevels[log["level"]], log["message"])
            # Uncomment this to print raw log-like data as a workaround, as described above.
            ##print(f"{log['level']} - {log['message']}")

    # Clear as much as possible, to present an almost-fresh instance of a browser for the next test. (Shutting down then starting up a browser is very slow.)
    driver.execute_script("window.localStorage.clear();")
    driver.execute_script("window.sessionStorage.clear();")
    driver.delete_all_cookies()


# _SeleniumUtils
# --------------
# Provide basic Selenium-based operations.
class _SeleniumUtils:
    def __init__(
        self,
        # These are fixtures.
        runestone_selenium_driver,
        host_address,
    ):

        self.driver = runestone_selenium_driver
        self.host_address = host_address
        self.wait = WebDriverWait(self.driver, 10)

    # A helper function to attach to the Selenium driver: get from a URL relative to the Runestone application.
    def get(self, relative_url):
        return self.driver.get("{}/{}".format(self.host_address, relative_url))

    # Scroll to the top of the window. A button can sometimes be scrolled to the top of the screen, where it's hidden by the navigation bar. In this case, we can't click it, since Selenium will complain ``Message: element click intercepted: Element <button class="btn btn-success run-button" type="button">...</button> is not clickable at point (460, 17). Other element would receive the click: <div class="navbar-collapse collapse navbar-ex1-collapse">...</div>``. To avoid this, scroll to the top of the document, guaranteeing that the navbar won't be hiding the run button.
    def scroll_to_top(self):
        self.driver.execute_script("window.scrollTo(0, 0);")

    # Wait until a Runestone component has finished rendering itself, given the ID of the component.
    def wait_until_ready(self, id):
        # The component is ready when it has the class below.
        self.wait.until(element_has_css_class((By.ID, id), "runestone-component-ready"))


# An expectation for Selenium, used for checking that an element has a particular css class. From the `Selenium docs <https://selenium-python.readthedocs.io/waits.html#explicit-waits>`_, under the "Custom wait conditions" subheading.
#
# locator - used to find the element
#
# returns the WebElement once it has the particular css class.
class element_has_css_class:
    def __init__(
        self,
        # The element to find; this is passed directly to `driver.find_element <https://selenium-python.readthedocs.io/api.html#selenium.webdriver.remote.webdriver.WebDriver.find_element>`_. See the `Selenium docs`_.
        locator,
        # The CSS class to look for.
        css_class,
    ):

        self.locator = locator
        self.css_class = css_class

    def __call__(self, driver):
        # Find the referenced element. Ignore stale elements.
        try:
            element = driver.find_element(*self.locator)
            if self.css_class in element.get_attribute("class"):
                return element
        except StaleElementReferenceException:
            pass
        return False



# ******************************************************************
# |docname| - Provide Sphinx support for fill-in-the-blank questions
# ******************************************************************
#
# .. Copyright (C) 2013  Bradley N. Miller
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "isaiahmayerchak"

import json
import ast
from numbers import Number
import re

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util import logging

from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    maybeAddToAssignment,
)
from runestone.common import (
    RunestoneIdDirective,
    RunestoneNode,
    RunestoneIdNode,
    get_node_line,
)
from runestone.common.xmlcommon import (
    substitute_departure,
    substitute_visitor,
)


def noop(self, node):
    pass


def setup(app):
    app.add_directive("fillintheblank", FillInTheBlank)
    app.add_role("blank", BlankRole)
    app.add_node(
        FITBNode,
        html=(visit_fitb_html, depart_fitb_html),
        xml=(visit_fitb_xml, depart_fitb_xml),
    )
    app.add_node(BlankNode, html=(visit_blank_html, depart_blank_html))
    app.add_node(
        FITBFeedbackNode,
        html=(visit_fitb_feedback_html, depart_fitb_feedback_html),
        xml=(noop, noop),
    )

    app.add_config_value("fitb_div_class", "runestone", "html")


class FITBNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


# Modified to mimic visit_fitb_feedback_html
def visit_fitb_html(self, node):
    # Original
    # node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])
    # self.body.append(node["delimiter"])
    #
    # res = node["template_start"] % node["runestone_options"]
    # self.body.append(res)
    # Modified
    # Save the HTML that's been generated so far. We want to know only what's generated inside this directive.
    self.context.append(self.body)
    self.body = []


def depart_fitb_html(self, node):
    # If there were fewer blanks than feedback items, add blanks at the end of the question. Also, determine the names of each blank.
    blank_names = {}
    blank_count = 0
    for blank_node in node.traverse(BlankNode):
        # Empty blanks have a "name" of ``-``.
        name = blank_node["input_name"]
        if name != "-":
            # Map from the blank's name to its index in the array of blanks. Don't include unnamed blanks.
            blank_names[name] = blank_count
        blank_count += 1
    while blank_count < len(node["feedbackArray"]):
        visit_blank_html(self, None)
        blank_count += 1

    # Warn if there are fewer feedback items than blanks.
    if len(node["feedbackArray"]) < blank_count:
        # Taken from the example in the `logging API <http://www.sphinx-doc.org/en/stable/extdev/logging.html#logging-api>`_.
        logger = logging.getLogger(__name__)
        logger.warning(
            "Not enough feedback for the number of blanks supplied.", location=node
        )

    # Capture HTML generated inside this directive.
    inner_html = self.body
    self.body = self.context.pop()

    # Generate the HTML. Start with required JSON data.
    db_dict = {
        "problemHtml": "".join(inner_html),
        "dyn_vars": node.dynamic,
        "blankNames": blank_names,
        "feedbackArray": node["feedbackArray"],
    }
    # Add in optional data.
    if dyn_imports := node["runestone_options"].get("dyn_imports", "").split():
        db_dict["dyn_imports"] = dyn_imports
    if (static_seed := node["runestone_options"].get("static_seed")) is not None:
        db_dict["static_seed"] = static_seed
    db_json = json.dumps(db_dict)

    # Some nodes (for example, those in a timed node) have their ``document == None``. Find a valid ``document``.
    node_with_document = node
    while not node_with_document.document:
        node_with_document = node_with_document.parent
    # Supply client-side grading info if we're not grading on the server.
    ssg = node_with_document.document.settings.env.config.runestone_server_side_grading
    if ssg:
        if node.dynamic:
            # Server-side graded dynamic problems render and provide the problem's HTML on the server; just tell the client it's a dynamic problem.
            client_json = json.dumps(dict(dyn_vars=True))
        else:
            # Other server-side graded problems need the problem's HTML.
            client_json = json.dumps(dict(problemHtml="".join(inner_html)))
    else:
        client_json = db_json
    node["runestone_options"]["client_json"] = client_json
    outer_html = (
        """
        <div class="%(divclass)s">
            <div data-component="fillintheblank" data-question_label="%(question_label)s" id="%(divid)s" %(optional)s style="visibility: hidden;">
                <script type="application/json">
                    %(client_json)s
                </script>
            </div>
        </div>
            """
        % node["runestone_options"]
    )

    # add HTML to the Database and clean up
    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        outer_html,
        db_json if ssg else None,
    )
    self.body.append(outer_html)


# <exercise>
#     <title>Fill-In, Integer Answer</title>

#     <statement>
#         <p>The game of bowling uses <var/> pins that you try to knock down.</p>
#     </statement>
#     <setup>
#         <var>
#             <condition number="10">
#                 <feedback>
#                     <p>Arranged in a triangle, there are <m>1+2+3+4 = 10</m> pins, a so-called <term>triangular</term> number.</p>
#                 </feedback>
#             </condition>
#             <condition number="16">
#                 <feedback>
#                     <p><em>Close</em>! You may have used hexadecimal notation, when you did not really mean to.</p>
#                 </feedback>
#             </condition>
#             <condition string=".*">
#                 <feedback>
#                     <p>Incorrect.</p>
#                 </feedback>
#             </condition>
#         </var>
#     </setup>
# </exercise>


def visit_fitb_xml(self, node):

    TEMPLATE_START = f"""
        <exercise label="{node["runestone_options"]["divid"]}">
            <statement>
    """
    self.output.append(TEMPLATE_START)
    node.private_children = node.children[1:]
    stmt = str(node.children[0])
    stmt = re.sub(r"<problematic.*?</problematic>", "<var />", stmt)
    neededBlanks = len(node["runestone_options"]["pattlist"])
    numBlanks = stmt.count("<var />")
    diff = neededBlanks - numBlanks
    if diff > 0:
        stmt = stmt[0 : stmt.rfind("</paragraph>")]
        for i in range(diff):
            stmt += " <var /> "
        stmt += " </paragraph>"
    node.children = []
    self.output.append(stmt)
    self.output.append("</statement>")


def depart_fitb_xml(self, node):

    # pattlist = node["runestone_options"]["pattlist"]  # answer patterns
    # flist = node["runestone_options"]["flist"]  # feedback
    # for _ in pattlist:
    #     blankCount += 1
    # while blankCount < len(flist):
    #     visit_blank_xml(self, None)
    #     blankCount += 1

    self.output.append("<setup>")
    # Walk the children of node
    # child 0 is the statement
    # children 1..N are the feedbacks
    #    each of these children will have an atribute called blankfeedbackdict which is the match for this
    #    each child will have a child/children that is the actual feedback we want to supply
    #

    for var in range(len(node["runestone_options"]["pattlist"])):
        self.output.append("<var>")

        for ans in range(len(node["runestone_options"]["pattlist"][var])):
            child = node.private_children.pop(0)
            if "blankfeedbackdict" in child:
                rx = child["blankfeedbackdict"]
                if "number" in rx:
                    self.output.append(f"""<condition number="{rx['number']}">""")
                else:
                    self.output.append(f"""<condition string="{rx['regex']}">""")
                fb = ""
                for p in child.children:
                    fb += str(p)
                self.output.append(f"<feedback>{fb}</feedback>")
                self.output.append("</condition>")
            else:
                self.output.append(str(child))
        self.output.append("</var>")
    self.output.append("</setup>")
    self.output.append("</exercise>")


class FillInTheBlank(RunestoneIdDirective):
    """
    .. fillintheblank:: some_unique_id_here

        Put the text of the question here.
        See https://runestone.academy/runestone/books/published/overview/Assessments/fitb.html
        for additional options and documentation.
        -   :Put the correct answer here: Put feedback displayed for this answer here.
            :x: Put feedback displayed for an incorrect answer here.
    """

    # config values (conf.py):
    #
    # - fitb_div_class - custom CSS class of the component's outermost div

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            # For dynamic problems, this contains JavaScript code which defines the variables used in template substitution in the problem. If this option isn't present, the problem will be a static problem.
            "dyn_vars": directives.unchanged,
            # For dynamic problems, this contain a space-separated list of libraries needed by this problem.
            "dyn_imports": directives.unchanged,
            # For dynamic problems, this provides an optional static seed.
            "static_seed": directives.unchanged,
            # case insensitive matching
            "casei": directives.flag,
        }
    )

    def run(self):
        """
        process the fillintheblank directive and generate html for output.
        :param self:
        :return: Nodes resulting from this directive.
        ...
        """

        super(FillInTheBlank, self).run()

        addQuestionToDB(self)

        fitbNode = FITBNode()
        fitbNode["runestone_options"] = self.options
        fitbNode["feedbackArray"] = []
        fitbNode["source"], fitbNode["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )

        self.updateContent()

        # Process dynamic problem content.
        env = self.state.document.settings.env
        dyn_vars = self.options.get("dyn_vars")
        # Store the dynamic code, or None if it's a static problem.
        fitbNode.dynamic = dyn_vars

        self.state.nested_parse(self.content, self.content_offset, fitbNode)
        self.options["divclass"] = env.config.fitb_div_class
        # Expected _`structure`, with assigned variable names and transformations made:
        #
        # .. code-block::
        #   :number-lines:
        #
        #   fitbNode = FITBNode()
        #       Item 1 of problem text
        #       ...
        #       Item n of problem text
        #       feedback_bullet_list = bullet_list()  <-- The last element in fitbNode.
        #           feedback_list_item = list_item()   <-- Feedback for the first blank.
        #               feedback_field_list = field_list()
        #                   feedback_field = field()
        #                       feedback_field_name = field_name()  <-- Contains an answer.
        #                       feedback_field_body = field_body()  <-- Contains feedback for this answer.
        #                   feedback_field = field()  <-- Another answer/feedback pair.
        #           feedback_list_item = bullet_item()  <-- Feedback for the second blank.
        #               ...etc. ...
        #
        # This becomes a data structure:
        #
        # .. code-block::
        #   :number-lines:
        #
        #   self.feedbackArray = [
        #       [   # blankArray
        #           { # blankFeedbackDict: feedback 1
        #               "regex" : feedback_field_name,      # (An answer, as a regex;
        #               "regexFlags" : "x",                 # "i" if ``:casei:`` was specified, otherwise "".) OR
        #               "number" : [min, max],              # a range of correct numeric answers OR
        #               "solution_code" : source_code,      # For dynamic problems -- an expression which evaluates
        #                                                   # to true or false to determine if the solution was correct.
        #               "feedback": feedback_field_body (after being rendered as HTML)  # Provides feedback for this answer.
        #           },
        #           { # Feedback 2
        #               Same as above.
        #           }
        #       ],
        #       [  # Blank 2, same as above.
        #       ],
        #       ...,
        #   ]
        #
        # ...and a transformed node structure:
        #
        # .. code-block::
        #   :number-lines:
        #
        #   fitbNode = FITBNode()
        #       Item 1 of problem text
        #       ...
        #       Item n of problem text
        #       FITBFeedbackNode(), which contains all the nodes in blank 1's feedback_field_body
        #       ...
        #       FITBFeedbackNode(), which contains all the nodes in blank n's feedback_field_body
        #
        self.assert_has_content()
        feedback_bullet_list = fitbNode.pop()
        if not isinstance(feedback_bullet_list, nodes.bullet_list):
            raise self.error(
                "On line {}, the last item in a fill-in-the-blank question must be a bulleted list.".format(
                    get_node_line(feedback_bullet_list)
                )
            )
        # The length of feedbback_list_item gives us the number of blanks.
        # the number of feedback is len(feedback_bullet_list.children[x].children[0].children)
        for feedback_list_item in feedback_bullet_list.children:
            assert isinstance(feedback_list_item, nodes.list_item)
            feedback_field_list = feedback_list_item[0]
            if len(feedback_list_item) != 1 or not isinstance(
                feedback_field_list, nodes.field_list
            ):
                raise self.error(
                    "On line {}, each list item in a fill-in-the-blank problems must contain only one item, a field list.".format(
                        get_node_line(feedback_list_item)
                    )
                )
            blankArray = []
            for feedback_field in feedback_field_list:
                assert isinstance(feedback_field, nodes.field)

                feedback_field_name = feedback_field[0]
                assert isinstance(feedback_field_name, nodes.field_name)
                feedback_field_name_raw = feedback_field_name.rawsource
                # Simply store the solution code for a dynamic problem.
                if dyn_vars:
                    blankFeedbackDict = {"solution_code": feedback_field_name_raw}
                else:
                    # See if this is a number, optionally followed by a tolerance.
                    try:
                        # Parse the number.
                        str_num, *list_tol = feedback_field_name_raw.split()
                        num = ast.literal_eval(str_num)
                        assert isinstance(num, Number)
                        # If no tolerance is given, use a tolerance of 0.
                        if len(list_tol) == 0:
                            tol = 0
                        else:
                            assert len(list_tol) == 1
                            tol = ast.literal_eval(list_tol[0])
                            assert isinstance(tol, Number)
                        # We have the number and a tolerance. Save that.
                        blankFeedbackDict = {"number": [num - tol, num + tol]}
                    except (SyntaxError, ValueError, AssertionError):
                        # We can't parse this as a number, so assume it's a regex.
                        regex = (
                            # The given regex must match the entire string, from the beginning (which may be preceded by whitespaces) ...
                            r"^\s*"
                            # ... to the contents (where a single space in the provided pattern is treated as one or more whitespaces in the student's answer) ...
                            + feedback_field_name.rawsource.replace(" ", r"\s+")
                            # ... to the end (also with optional spaces).
                            + r"\s*$"
                        )
                        blankFeedbackDict = {
                            "regex": regex,
                            "regexFlags": "i" if "casei" in self.options else "",
                        }
                        # Test out the regex to make sure it compiles without an error.
                        try:
                            re.compile(regex)
                        except Exception as ex:
                            raise self.error(
                                'Error when compiling regex "{}": {}.'.format(
                                    regex, str(ex)
                                )
                            )
                blankArray.append(blankFeedbackDict)
                feedback_field_body = feedback_field[1]
                assert isinstance(feedback_field_body, nodes.field_body)
                # Append feedback for this answer to the end of the fitbNode.
                ffn = FITBFeedbackNode(
                    feedback_field_body.rawsource,
                    *feedback_field_body.children,
                    **feedback_field_body.attributes,
                )
                ffn["blankFeedbackDict"] = blankFeedbackDict
                fitbNode += ffn

            # Add all the feedback for this blank to the feedbackArray.
            fitbNode["feedbackArray"].append(blankArray)

        maybeAddToAssignment(self)
        fitbNode["runestone_options"]["pattlist"] = fitbNode["feedbackArray"][:]
        fitbNode["runestone_options"]["flist"] = feedback_field_list[:]

        return [fitbNode]


# BlankRole
# ---------
# Create role representing the blank in a fill-in-the-blank question. This function returns a tuple of two values:
#
# 0. A list of nodes which will be inserted into the document tree at the point where the interpreted role was encountered (can be an empty list).
# #. A list of system messages, which will be inserted into the document tree immediately after the end of the current block (can also be empty).
def BlankRole(
    # _`roleName`: the local name of the interpreted role, the role name actually used in the document.
    roleName,
    # _`rawtext` is a string containing the enitre interpreted text input, including the role and markup. Return it as a problematic node linked to a system message if a problem is encountered.
    rawtext,
    # The interpreted _`text` content.
    text,
    # The line number (_`lineno`) where the interpreted text begins.
    lineno,
    # _`inliner` is the docutils.parsers.rst.states.Inliner object that called this function. It contains the several attributes useful for error reporting and document tree access.
    inliner,
    # A dictionary of directive _`options` for customization (from the "role" directive), to be interpreted by this function. Used for additional attributes for the generated elements and other functionality.
    options={},
    # A list of strings, the directive _`content` for customization (from the "role" directive). To be interpreted by the role function.
    content=[],
):

    # Blanks ignore all arguments, just inserting a blank.
    blank_node = BlankNode(rawtext, input_name=text)
    blank_node.line = lineno
    return [blank_node], []


class BlankNode(nodes.Inline, nodes.TextElement, RunestoneNode):
    pass


def visit_blank_html(self, node):
    # Note that the fitb visit code may call this with ``node = None``.
    name = node["input_name"] if node else ""
    # If the blank contained a name, use that as the name of the input element. A name of ``-`` (the default value for ``|blank|``, since there's no way to pass an empty value) is treated as an unnamed input element.
    html_name = "" if name == "-" else f" name={repr(name)}"
    self.body.append(f'<input type="text"{html_name} />')


def visit_blank_xml(self, node):
    self.output.append("<p><var /></p>")


def depart_blank_html(self, node):
    pass


# Contains feedback for one answer.
class FITBFeedbackNode(nodes.General, nodes.Element, RunestoneNode):
    pass


def visit_fitb_feedback_html(self, node):
    # Save the HTML generated thus far. Anything generated under this node will be placed in JSON.
    self.context.append(self.body)
    self.body = []


def depart_fitb_feedback_html(self, node):
    # Place all the HTML generated for this node and its children into the feedbackArray.
    node["blankFeedbackDict"]["feedback"] = "".join(self.body)
    # Restore HTML generated thus far.
    self.body = self.context.pop()



# .. Copyright (C) 2019 Bryan A. Jones.
#
# ******************************************************************************
# |docname| - A Runestone extension for authoring literate programming questions
# ******************************************************************************
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
import re
import os
from os import makedirs
from pathlib import Path
import json


# Third-party imports
# -------------------
import pygments
from pygments.lexers import guess_lexer_for_filename
from pygments.formatters import HtmlFormatter
from docutils import nodes
from docutils.transforms import Transform
from docutils.parsers.rst import directives
from docutils.parsers.rst.states import Struct
from CodeChat.CodeToRestSphinx import is_source_code
from CodeChat.CodeToRest import get_lexer


# Local imports
# -------------
from ..common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
    env_from_node,
)
from .lp_common_lib import STUDENT_SOURCE_PATH, code_here_comment, SPHINX_CONFIG_NAME
from ..server.componentdb import addQuestionToDB, addHTMLToDB


# Directives
# ----------
#
# Utilities
# ^^^^^^^^^
# _remove_code_solutions
# """"""""""""""""""""""
# Given code containing solution tags, remove the lines containing the solution tags and any intervening lines, replacing them with the provided string.
def _remove_code_solutions(
    # The file name of this code.
    file_name,
    # A string containing the code.
    src,
    # The _`replacement function`, invoked as ``replacement_func(start_line, end_line, file_name)`` which returns a replacement string.
    replacement_func,
):

    # Split these so they won't be treated as actual tags when this file is parsed.
    start_token = "SOLUTION_" "BEGIN"
    end_token = "SOLUTION_" "END"

    # Search through the lines of source from the end of the beginning. That way, any deletions made at the end won't change line numbers for deletions made earlier in the source.
    lines = src.splitlines(keepends=True)
    current_index = len(lines) - 1
    end_token_index = None
    while current_index >= 0:
        # Implement this is a two-state state machine:
        #
        # 1. When ``end_token_index`` is None, we're searching for the ending token.
        if end_token_index is None:
            if end_token in lines[current_index]:
                end_token_index = current_index
        # 2. Otherwise, ``end_token_index`` contains the line number at which the end token was found. Therefore, we're searching for the beginning token.
        else:
            if start_token in lines[current_index]:
                del lines[current_index : end_token_index + 1]
                lines.insert(
                    current_index,
                    replacement_func(current_index + 1, end_token_index + 1, file_name),
                )
                end_token_index = None
        current_index -= 1

    return "".join(lines)


# _assert_has_no_content
# """"""""""""""""""""""
# An almost-copy of ``docutils.parsers.rst.Directive.assert_has_content``. It throws an ERROR-level DirectiveError if the directive has contents.
def _assert_has_no_content(self):
    if self.content:
        raise self.error(
            'Content block not allowed for the "%s" directive.' % self.name
        )


# _source_read
# ^^^^^^^^^^^^
# The source-read_ event occurs when a source file is read. If it's code, this
# routine replaces code solutions with HTML textareas, so students can fill in their own code.
def _source_read(
    # .. _app:
    #
    # The `Sphinx application object <http://sphinx-doc.org/extdev/appapi.html#sphinx.application.Sphinx>`_.
    app,
    # The name of the document that was read. It contains a path relative to the
    # project directory and (typically) no extension.
    docname,
    # A list whose single element is the contents of the source file.
    source,
):

    if is_source_code(app.env, docname):
        source[0] = _remove_code_solutions(docname, source[0], _textarea_replacement)


# A `replacement function`_ to substitute a text area for student code for the solution.
def _textarea_replacement(start_line, end_line, file_name):

    # Create the minimal-length replacement string.
    s = TEXTAREA_REPLACEMENT_STRING.format(end_line - 4)

    # Pad with newlines if we can so line numbering is preserved.
    padding_newlines = end_line - start_line + 1 - s.count("\n")
    if padding_newlines > 0:
        s += "\n" * padding_newlines

    return s


TEXTAREA_REPLACEMENT_STRING = """
.. raw::
 html

 <textarea class="code_snippet"></textarea><br />

..

"""


# _LpBuildButtonDirective
# ^^^^^^^^^^^^^^^^^^^^^^^
class LpNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


def visit_lp_html(self, node):
    # Save the HTML that's been generated so far. We want to know only what's generated inside the LP directive.
    self.context.append(self.body)
    self.body = []


def depart_lp_html(self, node):
    env = env_from_node(node)
    id_ = node["runestone_options"]["divid"]
    html = """<div class="runestone">
    {}
    <input type="button" value="Save and run" class="btn btn-success" data-component="lp_build" data-question_label="%(question_label)s" data-lang="{}" id="{}" />
    <br />
    <textarea readonly class="lp-result"></textarea>
    <br />
    <div class="lp-feedback"><div></div></div>
</div>""".format(
        "".join(self.body), node["runestone_options"]["language"], id_
    )
    self.body = self.context.pop() + [html]

    addHTMLToDB(
        id_,
        node["runestone_options"]["basecourse"],
        html,
        json.dumps(
            dict(
                # Provide these to the server, indicating how to build.
                language=node["runestone_options"]["language"],
                builder=node["runestone_options"]["builder"],
                timelimit=node["runestone_options"]["timelimit"],
                include=node["runestone_options"]["include"],
                source_path=node["runestone_options"]["source_path"],
                sphinx_base_path=env.app.confdir,
            )
        ),
    )


# This inserts a "save and run" button, along with the HTML/JavaScript which causes the current page to be tested by compiling and running it with its test bench, then reporting the results to the user. It must only be used on pages that are a program and have a test bench associated with them. Only one should appear on a given page.
class _LpBuildButtonDirective(RunestoneIdDirective):
    # The required argument is an id_ for this question.
    required_arguments = 1
    # No optional arguments.
    optional_arguments = 0
    # Per http://docutils.sourceforge.net/docs/howto/rst-directives.html, True if content is allowed. However, this isn't checked or enforced.
    has_content = False
    # Options. Everything but language is currently ignored. This is based on activecode, so in the future similar support would be provided for these options.
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "include": directives.unchanged,
            "language": directives.unchanged,
            "timelimit": directives.unchanged,
            "stdin": directives.unchanged,
            "datafile": directives.unchanged,
            "available_files": directives.unchanged,
            "builder": directives.unchanged,
        }
    )

    def run(self):
        super(_LpBuildButtonDirective, self).run()
        _assert_has_no_content(self)
        addQuestionToDB(self)

        # Process options
        ##===============
        self.options["include"] = [x.strip() for x in self.options.get("include", "")]
        # If a language isn't provided, derive it from the file's name.
        env = self.state.document.settings.env
        self.options.setdefault(
            "language", get_lexer(filename=env.docname).mimetypes[0]
        )
        self.options.setdefault("timelimit", 25000)
        self.options.setdefault("builder", "JOBE")
        self.options["source_path"] = env.docname

        lp_node = LpNode()
        lp_node["runestone_options"] = self.options
        lp_node["source"], lp_node["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )

        return [lp_node]


# Remove code solutions and feedback from source files. Also, produce a Pygments-highlighted version of each source file.


def _doctree_resolved(app, doctree, docname):
    env = app.builder.env

    if is_source_code(env, docname):
        # Read the source of this file.
        src_path = env.doc2path(docname, None)
        src_abs_path = Path(env.srcdir) / src_path
        with src_abs_path.open(encoding="utf-8") as f_in:
            str_ = f_in.read()

        # See if any Runestone questions exist in this source file.
        # Remove feedback from source files.
        runestone_directives_to_remove = []
        for node in doctree.traverse(RunestoneIdNode):
            # See if any parent of this node is a RunestoneIdNode; if so, skip it.
            parent_node = node.parent
            while parent_node:
                if isinstance(parent_node, RunestoneIdNode):
                    break
                parent_node = parent_node.parent
            else:
                # None of the node's ancestors are RunestoneIdNodes. Add to our list.
                # Some nodes don't define enough information to remove them. Skip.
                if node["line"] and node.rawsource:
                    runestone_directives_to_remove += [
                        [node["line"], node["line"]
                            + len(node.rawsource.splitlines())]
                    ]
        if runestone_directives_to_remove:
            # Look for answers in the source and remove them. Work from the last answer (which is closest to the end of the file), so edits don't change the line numbering for earlier answers.
            l = str_.splitlines(keepends=True)
            for fb in reversed(runestone_directives_to_remove):
                del l[fb[0] : fb[1]]

            # Now that feedback is removed, convert it back to a string.
            str_ = "".join(l)

        # Replace solutions with "put your code here" comments. The lambda is a `replacement function`_.
        str_ = _remove_code_solutions(
            docname,
            str_,
            lambda start_line, end_line, file_name: code_here_comment(file_name),
        )

        # Write the source without answers or feedback into ``outdir/STUDENT_SOURCE_PATH``. Create the path if it doesn't exist.
        dest_path = Path(app.outdir) / STUDENT_SOURCE_PATH / src_path
        makedirs(str(dest_path.parent), exist_ok=True)
        with dest_path.open("w", encoding="utf-8") as f_out:
            f_out.write(str_)

        # _`Pygments source`: Write an HTML version of the source without answers or feedback using Pygments.
        #
        # Skip files if a lexer can't be found.
        try:
            # See http://pygments.org/docs/api/#module-pygments.
            lexer = guess_lexer_for_filename(docname, str_)
        except pygments.util.ClassNotFound:
            pass
        else:
            # Figure out the path from docname to the Pygments css file. For example, ``path/to/source/code.py`` becomes ``../../../_static/pygments.css``. The newer ``pathlib`` doesn't support this, while ``os.path.relpath`` does:
            #
            # .. code-block:: pycon
            #
            #   >>> os.path.relpath('/', '/a/b/c')
            #   '..\\..\\..'
            #   >>> Path('/').relative_to('/a/b/c')
            #   Traceback (most recent call last):
            #     File "<stdin>", line 1, in <module>
            #     File "E:\Downloads\Python36-64\lib\pathlib.py", line 873, in relative_to
            #       .format(str(self), str(formatted)))
            #   ValueError: '\\' does not start with '\\a\\b\\c'
            #
            # To make this work, prepend a ``/`` to both paths, signifying that they start from the same directory (app.outdir, in this case). Use the parent of docname. so that the file name isn't treated as a directory.
            cssfile = os.path.relpath(
                "/_static/pygments.css", "/" + str(Path(docname).parent)
            )
            # The directory containing the cssfile may not exist yet; create it if needed.
            makedirs(app.outdir + "/_static", exist_ok=True)
            # See http://pygments.org/docs/formatters/.
            formatter = HtmlFormatter(
                full=True,
                title=env.titles[docname],
                cssfile=cssfile,
                noclobber_cssfile=True,
            )
            pygments_name = Path(app.outdir) / (docname + "-source.html")
            # The directory containing pygments_name may not exist yet; create it if so.
            makedirs(str(pygments_name.parent), exist_ok=True)
            with pygments_name.open("w", encoding="utf-8") as f_out:
                # See http://pygments.org/docs/api/#module-pygments.
                pygments.highlight(str_, lexer, formatter, f_out)


#
# Anchored references
# ===================
# I have a number of rather long hyperlinks, and I want to refer to specific anchors of these hyperlinks. So, I'd like to provide a short name for the hyperlinks, then add an anchor to the end of the short name. This is :alink:`the finished product <short_name#external-links>`. Building up to this:
#
# .. _short_name: http://www.sphinx-doc.org/en/stable/rest.html
#
# Standard docutils syntax to refer to this is the usual `short_name`_ and the less well-known `this stuff <short_name_>`_. Per http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html#embedded-uris-and-aliases, this all works. But what I want is something like ``anchored stuff <short_name#anchor_>`_``. That doesn't natively work. While it seems tempting to play with Sphinx's `:ref: role <http://www.sphinx-doc.org/en/stable/markup/inline.html#role-ref>`_, this doesn't help, since ref doesn't refer to hyperlinks, but only labels. For example, ``:ref:`this link <short_name_>``` doesn't work.
#
# This functionality could also be replaced by the `sphinx.ext.extlinks <http://www.sphinx-doc.org/en/stable/ext/extlinks.html>`_ extension, which I found after finishing ``alink``. However, I prefer ``alink``: I don't like creating lots of new roles for every PDF I want to cross reference, and I like to be able to define a link locally in a document if it's used only there, rather than globally in the ``extlinks`` setting of `/conf.py`.
#
# To make this work:
#
# #. Create a new role, ``:alink:`title <refname#anchor>```. This role transforms its input to ```title <refname_>`_`` and has docutils parse it, adding the ``anchor`` to the resulting ``reference`` node produced by the parser.
# #. Create a transform which changes a ``refname`` into the corresponding ``refuri#anchor``. This process (without anchors) is performed by a `transform <http://docutils.sourceforge.net/docs/dev/hacking.html#transforming-the-document>`_.
# #. Register this role and transform with Sphinx.
#
# The _`:alink:` role
# --------------------
# Define the ``alink`` role, as discueed `above <anchored references>`_. `Docutils <http://docutils.sourceforge.net/docs/howto/rst-roles.html>`_ has documentation on this. There's a very helpful `tutorial <https://doughellmann.com/blog/2010/05/09/defining-custom-roles-in-sphinx/>`_ for creating a new role in Sphinx.
def _alink_role(
    # See roleName_.
    roleName,
    # See rawtext_.
    rawtext,
    # See text_.
    text,
    # See lineno_.
    lineno,
    # See inliner_.
    inliner,
    # See options_.
    options={},
    # See content_.
    content=[],
):

    # Look for ``title <refname#anchor>``.
    m = re.search(r"(.*\s+<[^#]+)(#.+)(>\s*)$", text)
    if not m:
        msg = inliner.reporter.error(
            'Expected "title <refname#anchor>", but saw "{}"'.format(text)
        )
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    anchor = m.group(2)
    # Reassemble this into a reference, but removing the anchor. It then becomes
    ## `title <refname_>`_
    no_anchor_reference = "`" + m.group(1) + "_" + m.group(3) + "`_"

    # Parse this. Booger a bit to be able to invoke the inliner. Just creating a reference node doesn't work, since the parser modifies internal data structures which relate refnames to references.
    memo = Struct(
        reporter=inliner.reporter, document=inliner.document, language=inliner.language
    )
    parsed_nodes, system_messages = inliner.parse(
        no_anchor_reference, lineno, memo, inliner.parent
    )

    # This should return [refence_node, target_node].
    assert isinstance(parsed_nodes[0], nodes.reference)
    assert isinstance(parsed_nodes[1], nodes.target)
    # Add in the anchor to the reference node.
    parsed_nodes[0]["anchor"] = anchor
    # I'd like ``:alink:`foo <bar#thing>``` followed by ``foo_`` to link to ``bar#thing``, but it instead links to ``bar``. That's because:
    #
    # #. ``:alink:`foo <bar#thing>``` is translated to ```foo <bar_>_```, an indirect reference to ``bar``.
    # #. In  (I think) ``docutils.transforms.references.IndirectHyperlinks``, the ``nodes.target`` referring to ``bar`` is replaced with the ``nodes.target`` for ``bar``. So, any info (such as an ``anchor`` field) specific to this target will be lost in the replacement.
    # #. In the ``ExternalAnchorTargets`` transform, the ``nodes.target`` with the actual ``refuri`` (name of the URL in a Docutils node) will have the ``node["reference"] anchor`` value appended. However, this will **only** apply to the original ``node["reference"]`` -- references to this reference won't have the ``anchor`` field. I'm not sure how or where to fix this (to make ``anchor`` propagate).
    #
    # Test cases: source link is
    # :alink:`.space directive <asmguide#page=75>`
    # and the following should refer to the same thing (but leaves off the anchor)
    # `.space directive`_.
    return parsed_nodes, system_messages


#
# Transform ``alink`` references
# ------------------------------
# This is based on ``docutils.transforms.references.ExternalTargets`` v0.12.
class ExternalAnchorTargets(Transform):

    """
    Given:

    .. code-block:: xml

        <paragraph>
            <reference refname="direct external#anchor">
                direct external
        <target id="id1" name="direct external" refuri="http://direct">

    The "refname" attribute is replaced by the direct "refuri" attribute with #anchor appended:

    .. code-block:: xml

        <paragraph>
            <reference refuri="http://direct#anchor">
                direct external
        <target id="id1" name="direct external" refuri="http://direct">
    """

    # This line was changed from the original. Lower is higher priority.
    # default_priority = 640
    default_priority = 639

    def apply(self):
        for target in self.document.traverse(nodes.target):
            if target.hasattr("refuri"):
                refuri = target["refuri"]
                for name in target["names"]:
                    reflist = self.document.refnames.get(name, [])
                    if reflist:
                        target.note_referenced_by(name=name)
                    for ref in reflist:
                        if ref.resolved:
                            continue
                        del ref["refname"]
                        # This line was changed from the original:
                        # ref['refuri'] = refuri
                        ref["refuri"] = refuri + ref.get("anchor", "")
                        ref.resolved = 1


#
# Hyperlinks to PDF documents
# ---------------------------
# The main use of the ``alink`` role is providing a hyperlink to a specific portion of a PDF document. Some reference material on this approach:
#
# #. You can't hyperlink to bookmarks in a PDF, but instead of named destinations.
# #. In terms of syntax, https://helpx.adobe.com/acrobat/kb/link-html-pdf-page-acrobat.html covers the basics. A longer, more complete guide is available from http://www.adobe.com/content/dam/Adobe/en/devnet/acrobat/pdfs/pdf_open_parameters.pdf.
# #. In some StackOverflow post, I read that using ``#nameddest=blah`` is a bit less suporrted than ``#blah``. Chrome seems to scroll to the page, but not the actual location.
# #. The utility `textify <http://rammichael.com/textify>`_ allows me to copy the name of a named destination, even on a locked PDF.
#
# Misc
# ====
# .. _`_docname_role`:
#
# _docname_role
# -------------
# Create a role which returns a specified part of the `docname <http://www.sphinx-doc.org/en/stable/extdev/envapi.html#sphinx.environment.BuildEnvironment.docname>`_ (the path to the current document). Syntax: ``:docname:`attr``` returns the ``attr`` method of the docname as a `Path <https://docs.python.org/3/library/pathlib.html#methods-and-properties>`_. For example, ``:docname:`name``` would return the name of the current docname_.
#
# This function returns a tuple of two values:
#
# 0. A list of nodes which will be inserted into the document tree at the point where the interpreted role was encountered (can be an empty list).
# #. A list of system messages, which will be inserted into the document tree immediately after the end of the current block (can also be empty).
def _docname_role(
    # _`roleName`: the local name of the interpreted role, the role name actually used in the document.
    roleName,
    # _`rawtext` is a string containing the enitre interpreted text input, including the role and markup. Return it as a problematic node linked to a system message if a problem is encountered.
    rawtext,
    # The interpreted _`text` content.
    text,
    # The line number (_`lineno`) where the interpreted text begins.
    lineno,
    # _`inliner` is the docutils.parsers.rst.states.Inliner object that called this function. It contains the several attributes useful for error reporting and document tree access.
    inliner,
    # A dictionary of directive _`options` for customization (from the "role" directive), to be interpreted by this function. Used for additional attributes for the generated elements and other functionality.
    options={},
    # A list of strings, the directive _`content` for customization (from the "role" directive). To be interpreted by the role function.
    content=[],
):

    # See https://doughellmann.com/blog/2010/05/09/defining-custom-roles-in-sphinx/.
    env = inliner.document.settings.env
    try:
        # Invoke
        p = Path(env.docname)
        # Return p.<text> using `getattr <https://docs.python.org/3/library/functions.html#getattr>`_.
        path_component = str(getattr(p, text, p))
    except Exception as e:
        # Report an error.
        msg = inliner.reporter.error(
            "Invalid path component {}: {}".format(text, e), line=lineno
        )
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    # Create a reference to the underlying source file. Typical node structure:
    #
    # .. code-block:: guess
    #
    #   <reference name="link text" refuri="../link.html">
    #       link text
    #   </reference>
    #   <target ids="['link-text']" names="['link text']" refuri="../link.html"/>
    #
    # For simplicity, I omit the target and just insert the reference. The source HTML is created from `Pygments source`_.
    refuri = str(Path(env.docname).name) + "-source.html"
    return [nodes.reference(rawtext, path_component, refuri=refuri, **options)], []


#
# setup
# =====
# Register these additions with Sphinx.
def setup(
    # Sphinx application context.
    app,
):

    # This depends on `CodeChat <https://pythonhosted.org/CodeChat/README.html>`_.
    app.setup_extension("CodeChat.CodeToRestSphinx")

    # Supply a fake CSS file to avoid errors, since the CodeChat's CSS will import this.
    app.add_autoversioned_javascript("html4css1.css", defer="")

    # See http://www.sphinx-doc.org/en/stable/extdev/appapi.html#sphinx.application.Sphinx.add_role.
    app.add_role("alink", _alink_role)
    app.add_role("docname", _docname_role)

    # See http://www.sphinx-doc.org/en/stable/extdev/appapi.html#sphinx.application.Sphinx.add_transform.
    app.add_transform(ExternalAnchorTargets)

    app.add_node(LpNode, html=(visit_lp_html, depart_lp_html))

    # See http://www.sphinx-doc.org/en/stable/extdev/appapi.html#sphinx.application.Sphinx.add_directive.
    app.add_directive("lp_build", _LpBuildButtonDirective)

    # Use the `source-read <http://sphinx-doc.org/extdev/appapi.html#event-source-read>`_
    # event hook to replace solutions with textareas before Sphinx processes it.
    app.connect("source-read", _source_read)
    # See http://www.sphinx-doc.org/en/stable/extdev/appapi.html#event-doctree-resolved.
    app.connect("doctree-resolved", _doctree_resolved)

    # Save Sphinx build info for use in other applications (web server, build system, etc.)
    with open(SPHINX_CONFIG_NAME, "w", encoding="utf-8") as f:
        f.write(
            json.dumps(
                {
                    "SPHINX_SOURCE_PATH": str(Path(app.srcdir).relative_to(Path.cwd())),
                    "SPHINX_OUT_PATH": str(Path(app.outdir).relative_to(Path.cwd())),
                }
            )
        )

    # Return `extension metadata <http://sphinx-doc.org/extdev/index.html>`_.
    return {"version": "0.0.1", "parallel_read_safe": True}



#!/usr/bin/env python3
#
# `build.py` - Build the wheels and Docker containers needed for this
# application
# ===================================================================
#
# Imports
# -------
import locale
import os
import subprocess
import sys
from shutil import copyfile
import yaml
import toml
import pdb  # noqa
from collections import OrderedDict
from datetime import datetime

# use python-dotenv >= 0.21.0
from dotenv import load_dotenv
from rich.console import Console
from rich.live import Live
from rich.table import Table
from sqlalchemy import create_engine, inspect
import click
from rsptx.cl_utils.core import pushd


class Config(object):
    def __init__(self):
        self.verbose = False


pass_config = click.make_pass_decorator(Config, ensure=True)

console = Console()


@click.group(chain=True)
@click.option("--verbose", is_flag=True, help="Show more output")
@click.option(
    "--all",
    is_flag=True,
    show_default=True,
    help="Build all containers, including author and worker",
    default=False,
)
@click.option(
    "--core",
    is_flag=True,
    show_default=True,
    help="Build only the core services",
    default=False,
)
@click.option("-s", "--service", multiple=True, help="Build one service - multiple ok")
@click.option(
    "--clean", is_flag=True, help="Remove all containers and images before starting"
)
@pass_config
def cli(config, verbose, all, core, service, clean):
    """
    Build the wheels and Docker containers needed for this application
    You can control which parts of the pipeline to run and which services to build or
    restart.  The default is to build the core services and restart them.  Here are some
    examples:
    build --core full
    build wheel image restart
    build wheel image push
    build --service author --service worker wheel image restart
    build push -- push the built images to the registry
    """

    os.chdir(findProjectRoot())

    if not os.path.exists(".env"):
        console.print(
            "No .env file found.  Please copy sample.env to .env and edit it.",
            style="bold red",
        )
        exit(1)
    config.verbose = verbose

    if clean:
        clean_all()

    if not service and not all:
        click.echo("No services specified, building core services")
        core = True

    with open("pyproject.toml") as f:
        config.pyproject = toml.load(f)
        config.version = config.pyproject["tool"]["poetry"]["version"]

    res = subprocess.run("docker info", shell=True, capture_output=True)
    if res.returncode != 0:
        console.print(
            "Docker is not running.  Please start it and try again.", style="bold red"
        )
        exit(1)

    # make a fresh build.log for this build
    with open("build.log", "w") as f:
        f.write("")

    ym = yaml.load(open("docker-compose.yml"), yaml.FullLoader)

    config.orig = ym.copy()

    # remove the redis service from the list since we don't customize it
    del ym["services"]["redis"]

    # add the interactives service to the list so it gets built before the others,
    # assignments depend on it.  We do not build a container for it, but we do build the wheels
    ym["services"]["interactives"] = {"build": {"context": "./projects/interactives"}}
    ym["services"]["interactives"]["image"] = "skip"
    ym["services"] = OrderedDict(ym["services"])  # convert to ordered dict
    ym["services"].move_to_end("interactives", last=False)

    # remove all services except the one we want to build
    if service:
        core = False
        svc_list = [x for x in ym["services"].keys()]
        for svc in svc_list:
            if svc not in service:
                del ym["services"][svc]

    if core:
        svc_list = [x for x in ym["services"].keys()]
        to_remove = [
            "author",
            "worker",
            "nginx_dstart_dev",
            "pgbouncer",
            "interactives",
            "db",
        ]
        if not os.path.exists(
            "bases/rsptx/interactives/runestone/dist/webpack_static_imports.json"
        ):
            to_remove.remove("interactives")
            click.echo("Keeping interactives in the build list since they are missing")
        for svc in svc_list:
            if svc in to_remove:
                del ym["services"][svc]
    # Now initialize the build.log files for all services
    for service in ym["services"]:
        now = datetime.now()
        try:
            _ = ym["services"][service]["build"]["context"]
        except KeyError:
            continue
        with pushd(ym["services"][service]["build"]["context"]):
            with open("build.log", "w") as f:
                f.write(f"Build log for {service} on {now}\n")

    service_list = ", ".join(ym["services"].keys())
    console.print(f"Building services: {service_list}", style="bold")

    config.all = all
    config.core = core
    config.ym = ym


def clean_all():
    console.print("Removing all containers and images...", style="bold")
    ret = subprocess.run(["docker", "system", "prune", "--force"], capture_output=True)
    if ret.returncode == 0:
        console.print("All containers and images removed successfully", style="green")
    else:
        console.print("Failed to remove all containers and images", style="bold red")
        exit(1)


def findProjectRoot():
    start = os.getcwd()
    prevdir = ""
    while start != prevdir:
        if os.path.exists(os.path.join(start, "docker-compose.yml")):
            if prevdir != "":
                console.print(
                    f"Changing working directory to {start}", style="bold blue"
                )
            return start
        prevdir = start
        start = os.path.dirname(start)
    raise IOError("You must be in the rs repo to run the build command")


@cli.command()
@pass_config
def list(config):
    """List the services and last build"""

    table = Table(title="Services")
    table.add_column("Service", justify="right", style="grey62", no_wrap=True)
    table.add_column("Last Built", style="magenta")

    for s in config.orig["services"]:
        # get the mtime of build.log in project/service and print it
        try:
            projdir = config.orig["services"][s]["build"]["context"]
        except KeyError:
            mtime = "No Context"
            continue
        with pushd(projdir):
            try:
                mtime = os.path.getmtime("build.log")
                mtime = datetime.fromtimestamp(mtime)
            except FileNotFoundError:
                mtime = "Not Built"
            #            console.print(f"{s} - {mtime}")
            table.add_row(s, str(mtime))
    console.print(table)


@cli.command()
@pass_config
def env(config):
    """Check key environment variables and exit"""

    console.print("Checking environment variables", style="bold")

    # Per the [docs](https://pypi.org/project/python-dotenv/), load `.env` into
    # environment variables.
    load_dotenv()
    table = Table(title="Environment Variables")
    table.add_column("Variable", justify="right", style="grey62", no_wrap=True)
    if config.verbose:
        table.add_column("Value", style="magenta")
    else:
        table.add_column("Set", style="magenta")
    finish = False
    for var in [
        "RUNESTONE_PATH",
        "RUNESTONE_HOST",
        "SERVER_CONFIG",
        "BOOK_PATH",
        "WEB2PY_CONFIG",
        "JWT_SECRET",
        "WEB2PY_PRIVATE_KEY",
        "COMPOSE_PROFILES",
        "DBURL",
        "DEV_DBURL",
    ]:
        if var not in os.environ:
            table.add_row(var, "[red]No[/red]")
            finish = True
        else:
            if config.verbose:
                table.add_row(var, os.environ[var])
            else:
                table.add_row(var, "[green]Yes[/green]")

    if "DC_DBURL" not in os.environ:
        table.add_row("DC_DBURL", "[red]No[/red] will default to DBURL")
        if "DBURL" not in os.environ:
            table.add_row("DBURL", "[red]No[/red] please set it in .env")
            finish = True
    else:
        if config.verbose:
            table.add_row("DC_DBURL", os.environ["DC_DBURL"])
        else:
            table.add_row("DC_DBURL", "[green]Yes[/green]")

    if "DC_DEV_DBURL" not in os.environ:
        table.add_row("DC_DEV_DBURL", "[red]No[/red] will default to DEV_DBURL")
        if "DEV_DBURL" not in os.environ:
            table.add_row("DEV_DBURL", "[red]No[/red] please set it in .env")
            finish = True
    else:
        if config.verbose:
            table.add_row("DC_DEV_DBURL", os.environ["DC_DEV_DBURL"])
        else:
            table.add_row("DC_DEV_DBURL", "[green]Yes[/green]")

    console.print(table)

    cprofs = os.environ.get("COMPOSE_PROFILES", "")
    if "basic" in cprofs:
        console.print("Basic profile enabled")
        if "2345" not in os.environ.get("DEV_DBURL", ""):
            console.print(
                "DEV_DBURL is not using port 2345.  This is required to connect to a dockerized postgres instance.  Please set it in .env",
                style="bold red",
            )
            finish = True
    else:
        console.print(
            "You are configured to run your own DB instance.  Make sure it is running."
        )
        console.print(
            "If you don't want to run your own, then enable the basic profile."
        )
        if "2345" in os.environ.get("DEV_DBURL", ""):
            console.print(
                "DEV_DBURL is using port 2345.  This is probably not correct if you are running postgres outside of docker",
                style="bold red",
            )
            finish = True

    if "COMPOSE_PROFILES" not in os.environ:
        console.print(
            "COMPOSE_PROFILES not set.  Options include: basic, author, dev, production",
            style="bold red",
        )
        console.print(
            """
            The following services [red]will not be started[/red] unless you use set your COMPOSE_PROFILES in .env
            [green]db, author, worker pgbouncer, nginx_dstart_dev[/green]
            See the sample.env file for more information to set this variable.
            You can also use the --profile switch with docker compose to enable a specific profile.
            """,
            style="bold",
        )
        finish = False

    if not os.path.isfile(
        "bases/rsptx/web2py_server/applications/runestone/models/1.py"
    ):
        console.print("Copying 1.py.prototype to 1.py")
        copyfile(
            "bases/rsptx/web2py_server/applications/runestone/models/1.py.prototype",
            "bases/rsptx/web2py_server/applications/runestone/models/1.py",
        )

    if finish:
        console.print(
            "Your environment is not set up correctly.  Please define the environment variables listed above.",
            style="bold red",
        )
        exit(1)


# Attempt to determine the encoding of data returned from stdout/stderr of
# subprocesses. This is non-trivial. See the discussion at [Python's
# sys.stdout](https://docs.python.org/3/library/sys.html#sys.stdout). First, try
# `locale.getencoding()` (requires >= 3.11), with a fallback to
# `locale.getpreferredencoding()`.
#
# This is motivated by errors on Windows under Python 3.10:
#
#       Traceback (most recent call last):
#       File "C:\Users\bjones\Documents\git\rs\build.py", line 94, in <module>
#           f.write(res.stdout.decode("utf-8"))
#       File "C:\Users\bjones\AppData\Local\Programs\Python\Python310\lib\encodings\cp1252.py", line 19, in encode
#           return codecs.charmap_encode(input,self.errors,encoding_table)[0]
#       UnicodeEncodeError: 'charmap' codec can't encode characters in position 55319-55358: character maps to <undefined>
try:
    # Since this isn't defined on all Python versions, tell mypy to ignore the
    # following error.
    stdout_err_encoding = locale.getencoding()  # type: ignore[attr-defined]
except AttributeError:
    stdout_err_encoding = locale.getpreferredencoding()


def progress_wheel():
    chars = "x+"
    progress_wheel.counter += 1
    return chars[progress_wheel.counter % 2]


progress_wheel.counter = 0


def generate_wheel_table(status: dict) -> Table:
    table = Table(title="Build Python Wheels")
    table.add_column("Wheel", justify="right", style="white", no_wrap=True)
    table.add_column("Built", style="magenta")
    for service in status:
        table.add_row(service, status[service])
    return table


@cli.command()
@pass_config
def wheel(config):
    """Build the python wheels"""
    status = {}
    with Live(generate_wheel_table(status), refresh_per_second=4) as lt:
        status = {}
        for proj in config.ym["services"].keys():
            if (
                "build" not in config.ym["services"][proj]
                or config.ym["services"][proj]["build"]["context"] == "./"
            ):
                status[proj] = "[blue]Skipped[/blue]"
                lt.update(generate_wheel_table(status))
                continue
            projdir = config.ym["services"][proj]["build"]["context"]
            if os.path.isdir(projdir):
                with pushd(projdir):
                    if os.path.isfile("build.py"):
                        status[proj] = "[grey62]pre build...[/grey62]"
                        lt.update(generate_wheel_table(status))
                        res = subprocess.run(
                            ["python", "build.py", "--fromroot"], capture_output=True
                        )
                        if res.returncode == 0:
                            status[proj] = "[green]Yes[/green]"
                            lt.update(generate_wheel_table(status))
                            with open("build.log", "a") as f:
                                f.write(res.stdout.decode(stdout_err_encoding))
                        else:
                            status[proj] = f"[red]Fail[/red] See {projdir}/build.log"
                            lt.update(generate_wheel_table(status))
                            if config.verbose:
                                console.print(res.stderr.decode(stdout_err_encoding))
                            else:
                                with open("build.log", "a") as f:
                                    f.write(res.stderr.decode(stdout_err_encoding))
                                    f.write(res.stdout.decode(stdout_err_encoding))
                            sys.exit(1)
                    if os.path.isfile("pyproject.toml"):
                        status[proj] = "[grey62]building...[/grey62]"
                        lt.update(generate_wheel_table(status))
                        toml_stat = os.stat("pyproject.toml")
                        lock_stat = os.stat("poetry.lock")
                        if toml_stat.st_mtime > lock_stat.st_mtime:
                            res = subprocess.run(
                                ["poetry", "--version"], capture_output=True
                            )
                            # poetry 2.0 removed the --no-update flag (it is the default)
                            lock_opts = ["poetry", "lock"]
                            if "version 1" in res.stdout.decode(stdout_err_encoding):
                                lock_opts.append("--no-update")

                            res = subprocess.run(lock_opts, capture_output=True)
                            if res.returncode != 0:
                                status[proj] = (
                                    f"[red]Fail[/red] probable dependency conflict see {projdir}/build.log"
                                )
                                lt.update(generate_wheel_table(status))
                                if config.verbose:
                                    console.print(
                                        res.stderr.decode(stdout_err_encoding)
                                    )
                                else:
                                    with open("build.log", "a") as f:
                                        f.write(res.stderr.decode(stdout_err_encoding))
                                sys.exit(1)
                        res = subprocess.run(
                            ["poetry", "build-project"], capture_output=True
                        )
                        if res.returncode == 0:
                            status[proj] = "[green]Yes[/green]"
                            lt.update(generate_wheel_table(status))
                            with open("build.log", "a") as f:
                                f.write(res.stdout.decode(stdout_err_encoding))
                        else:
                            status[proj] = f"[red]Fail[/red] see {projdir}/build.log"
                            lt.update(generate_wheel_table(status))
                            if config.verbose:
                                console.print(res.stderr.decode(stdout_err_encoding))
                            else:
                                with open("build.log", "a") as f:
                                    f.write(res.stderr.decode(stdout_err_encoding))
                            sys.exit(1)
                    else:
                        status[proj] = "[blue]Skipped[/blue]"
                        lt.update(generate_wheel_table(status))


# Generate a table for the Live object
# see https://rich.readthedocs.io/en/stable/live.html?highlight=update#basic-usage
def generate_table(status: dict) -> Table:
    table = Table(title="Build Docker Images")
    table.add_column("Service", justify="right", style="white", no_wrap=True)
    table.add_column("Built", style="magenta")
    for service in status:
        table.add_row(f"{service}", status[service])
    return table


@cli.command()
@pass_config
def image(config):
    """Build the docker images"""

    console.print(
        "Building docker images (see build.log for detailed progress)...", style="bold"
    )
    status = {}
    with Live(generate_table(status), refresh_per_second=4) as lt:
        status = {}
        for service in config.ym["services"]:
            if service == "interactives":
                status[service] = "[grey62]skipped...[/grey62]"
                lt.update(generate_table(status))
                continue
            status[service] = "[grey62]building...[/grey62]"
            lt.update(generate_table(status))
            # to use a different wheel without editing Dockerfile use --build-arg wheel="wheelname"
            command_list = [
                "docker",
                "compose",
                "-f",
                "docker-compose.yml",
                "--progress",
                "plain",
                "build",
                service,
            ]

            with open("build.log", "ab") as f:
                ret = subprocess.run(
                    command_list,
                    capture_output=True,
                )
                f.write(ret.stdout)
                f.write(ret.stderr)
            if ret.returncode == 0:
                status[service] = "[green]Yes[/green]"
                lt.update(generate_table(status))
            else:
                status[service] = "Failed"
                lt.update(generate_table(status))
                console.print(
                    f"There was an error building {service} see build.log for details",
                    style="bold red",
                )
                exit(1)


# Now if the --push flag was given, push the images to Docker Hub
# For this next part to work you need to be logged in to a docker hub account or the
# runestone private registry on digital ocean.  The docker-compose.yml file has the
# image names set up to push to the runestone registry on digital ocean.
@cli.command()
@pass_config
def push(config):
    """Push the images to Docker Hub"""
    try:
        with open(".last_version") as f:
            last_version = f.read().strip()
    except FileNotFoundError:
        with open(".last_version", "w") as f:
            f.write(config.version)
            last_version = config.version

    if last_version == config.version:
        update_version = input(
            f"Do you want to update the version number ({config.version}) in pyproject.toml?"
        )
        if update_version.lower() in ["y", "yes"]:
            new_version = input("Enter the new version number: ")
            subprocess.run(
                ["poetry", "version", new_version], capture_output=True, check=True
            )
            console.out("Version updated, don't forget to commit the change.")
            # add a git tag for the new version
            subprocess.run(
                ["git", "tag", f"v{new_version}"], capture_output=True, check=True
            )
            subprocess.run(
                ["git", "push", "--tags"], capture_output=True, check=True
            )
            config.version = new_version

    console.print("Pushing docker images to Docker Hub...", style="bold")
    for service in config.ym["services"]:
        if "image" in config.ym["services"][service]:
            image = config.ym["services"][service]["image"]
            if "ghcr.io" not in image:
                continue
            console.print(f"Pushing {image}")
            ret1 = subprocess.run(
                ["docker", "tag", image, f"{image}:v{config.version}"], check=True
            )
            # we do 2 pushes because the version tag is not automatically pushed
            # and we want to push both the latest and the version tagged images
            # but it is low cost as we only push the layers that are not already
            # on the server.
            ret2 = subprocess.run(["docker", "push", "--quiet", image], check=True)
            ret3 = subprocess.run(
                ["docker", "push", "--quiet", f"{image}:v{config.version}"], check=True
            )

            if ret1.returncode + ret2.returncode + ret3.returncode == 0:
                console.print(f"{image} pushed successfully")
            else:
                console.print(f"{image} failed to push", style="bold red")
                exit(1)

    console.print("Docker images pushed successfully", style="green")
    with open(".last_version", "w") as f:
        f.write(config.version)


@cli.command()
@pass_config
def checkdb(config):
    """Check the database and run migrations"""
    # Check the database
    # ------------------
    engine = create_engine(os.environ["DEV_DBURL"])
    try:
        inspector = inspect(engine)
        _ = engine.connect()
    except Exception as e:
        console.print(f"Failed to connect to the database: {e}", style="bold red")
        exit(1)
    DBOK = True
    # Check if the alembic_version table is present in the database
    try:
        inspector.get_table_names().index("useinfo")
    except Exception:
        console.print(
            "It appears your database has not been initialized or your database is not started. Please see the documentation for how to initialize the database.",
            style="bold red",
        )
        DBOK = False
    try:
        inspector.get_table_names().index("alembic_version")
    except Exception:
        console.print(
            "The Migrations have not been set up for your database.  You need to run the alembic stamp head command to create it.",
            style="bold red",
        )
        DBOK = False

    if DBOK:
        console.print("Database appears to be set up correctly", style="green")
        console.print("Will now check for any needed migrations")
        res = subprocess.run(["alembic", "heads"], check=True, capture_output=True)
        head = res.stdout.decode("utf-8").split()[0].strip()
        console.print(f"Alembic head is {head}")
        res = subprocess.run(["alembic", "current"], check=True, capture_output=True)
        current = res.stdout.decode("utf-8").split("\n")[-2].split()[0].strip()
        console.print(f"Alembic current is {current}")
        if res.returncode == 0:
            if head == current:
                console.print(
                    "Your database schema appears to be up to date", style="green"
                )
            else:
                console.print("Database migrations are needed.  Running them now...")
                res = subprocess.run(["alembic", "upgrade", "head"], check=True)
                if res.returncode == 0:
                    console.print("Migrations completed successfully", style="green")
                else:
                    console.print("Migrations failed", style="bold red")
                    res = subprocess.run(
                        ["alembic", "upgrade", "--sql", "head"], capture_output=True
                    )
                    console.print(
                        "Try running the following commands manually in psql to see what went wrong."
                    )
                    console.print(
                        "Then run [green]alembic stamp head[/green] to mark the migrations as complete."
                    )
                    console.print(res.stdout.decode("utf-8"))

    if not DBOK:
        exit(1)


@cli.command()
@pass_config
def restart(config):
    """Restart the runestone docker services"""
    profiles = []
    profiles += [prof for prof in os.environ.get("COMPOSE_PROFILES", "").split(",")]
    profiles = [prof for prof in profiles if prof]
    if config.all or config.core:
        console.print("Restarting all services...", style="bold")
        if config.all and "author" not in profiles:
            profiles.append("author")
            os.environ["COMPOSE_PROFILES"] = ",".join(profiles)
        command_list = [
            "docker",
            "compose",
            "-f",
            "docker-compose.yml",
            "stop",
        ]
        ret1 = subprocess.run(
            command_list,
            capture_output=True,
        )
        if ret1.returncode != 0:
            console.print("Runestone services failed to stop", style="bold red")
            if config.verbose:
                console.print(ret1.stderr.decode("utf-8"))
            exit(1)
    else:
        for service in config.ym["services"]:
            command_list = [
                "docker",
                "compose",
                "-f",
                "docker-compose.yml",
                "stop",
                service,
            ]
            console.print(f"Restarting the {service} service...", style="bold")
            ret1 = subprocess.run(
                command_list,
                capture_output=True,
            )
            if ret1.returncode != 0:
                console.print(
                    f"Runestone service {service} failed to stop", style="bold red"
                )
                exit(1)
    if config.all or config.core:
        command_list = [
            "docker",
            "compose",
            "-f",
            "docker-compose.yml",
            "up",
            "-d",
        ]
        ret2 = subprocess.run(
            command_list,
            capture_output=True,
        )
        if ret2.returncode != 0:
            console.print("Runestone services failed to restart", style="bold red")
            exit(1)
    else:
        for service in config.ym["services"]:
            command_list = [
                "docker",
                "compose",
                "-f",
                "docker-compose.yml",
                "up",
                "-d",
                service,
            ]
            ret2 = subprocess.run(
                command_list,
                capture_output=True,
            )
            if ret2.returncode != 0:
                console.print(
                    f"Runestone service {service} failed to restart", style="bold red"
                )
                exit(1)
    console.print("Runestone services restarted successfully", style="green")


# This is a cool trick with click that lets you chain commands together to
# form a meta command. so this will run the env command first, then the wheel
# command, then the image command, then checkdb, then the restart command.
@cli.command()
@pass_config
@click.pass_context
def full(ctx, config):
    """Check the environment, Build the wheels, images, and run migrations, then restart the services"""
    ctx.invoke(env)
    ctx.invoke(wheel)
    ctx.invoke(image)
    ctx.invoke(checkdb)
    ctx.invoke(restart)


# This is a cool trick with click that lets you chain commands together to
# form a meta command. so this will run the env command first, then the wheel
# command, then the image command, then checkdb, then the restart command.
@cli.command()
@pass_config
@click.pass_context
def dev(ctx, config):
    """Build the wheels, images, and restart the services"""
    ctx.invoke(wheel)
    ctx.invoke(image)
    ctx.invoke(restart)


if __name__ == "__main__":
    cli()



# *********
# |docname|
# *********
# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "bmiller"

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import RunestoneIdDirective


def setup(app):
    app.add_directive("animation", Animation)
    app.add_autoversioned_javascript("animationbase.js", defer="")


SRC = """
<div id="%(divid)s" class="%(optclass)s">
<canvas id="%(divid)s_canvas" width="400" height="400" style="border:4px solid blue"></canvas>
<br />
<button onclick="%(divid)s_anim = %(divid)s_init('%(divid)s')">Initialize</button>
<button onclick="%(divid)s_anim.run('%(divid)s_anim')">Run</button>
<button onclick="%(divid)s_anim.stop()">Stop</button> </br>
<button onclick="%(divid)s_anim.begin()">Beginning</button>
<button onclick="%(divid)s_anim.forward()">Step Forward</button>
<button onclick="%(divid)s_anim.backward()">Step Backward</button>
<button onclick="%(divid)s_anim.end()">End</button>

<script>
%(divid)s_init = function(divid)
{
   var a = new Animator(new %(model)s(), new %(viewer)s(), divid)
   a.init()
   return a
}
</script>

</div>
"""

SCRIPTTAG = """<script src="../_static/%s"></script>\n"""


class Animation(RunestoneIdDirective):
    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True
    has_content = False
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "modelfile": directives.unchanged,
            "viewerfile": directives.unchanged,
            "model": directives.unchanged,
            "viewer": directives.unchanged,
        }
    )

    def run(self):
        """
        process the video directive and generate html for output.
        :param self:
        :return:
        """
        super(Animation, self).run()
        res = ""

        if "modelfile" in self.options:
            res = res + SCRIPTTAG % self.options["modelfile"]
        if "viewerfile" in self.options:
            res = res + SCRIPTTAG % self.options["viewerfile"]

        res = res + SRC % self.options
        rawnode = nodes.raw(self.block_text, res, format="html")
        rawnode.source, rawnode.line = self.state_machine.get_source_and_line(
            self.lineno
        )
        return [rawnode]


source = """
.. animation:: testanim
   :modelfile: sortmodels.js
   :viewerfile: sortviewers.js
   :model: SortModel
   :viewer: BarViewer

"""

if __name__ == "__main__":
    from docutils.core import publish_parts

    directives.register_directive("animation", Animation)

    doc_parts = publish_parts(
        source,
        settings_overrides={"output_encoding": "utf8", "initial_header_level": 2},
        writer_name="html",
    )

    print(doc_parts["html_body"])



# *********************************
# |docname| - An Group Submissions
# *********************************
#
# This directive lets you specify a question by random selection
# Given a list of question ids, it will randomly select one of those ids
# to present to the student.
# given a competency it will select a random question from all questions that
# test for that competency.

# Copyright (C) 2021  Runestone Academy LLC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = "bmiller"

# Imports from standard libarary
# ------------------------------

# Imports from third party libraries
# ----------------------------------
from docutils import nodes
from docutils.parsers.rst import directives

# local imports
# -------------
from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    maybeAddToAssignment,
)
from runestone.common.runestonedirective import (
    RunestoneDirective,
)


TEMPLATE = """
<div class="runestone sqcontainer %(optclass)s">
    <div data-component="groupsub" id={component_id} {size_limit}>
        <div class="col-sm-6">
            <select id="assignment_group" multiple class="assignment_partner_select" style="width: 100%">
            </select>
        </div>
        <div id="groupsub_button" class="col-sm-6">
        </div>
        <p>The Submit Group button will submit the answer for each each question
        on this page for each member of your group. It also logs you as the official
        group submitter.</p>
    </div>
</div>
"""


def setup(app):
    app.add_directive("groupsub", GroupSubmission)


class GroupSubmission(RunestoneDirective):
    """
    .. groupsub:: uniqueid
       :optional:
       :limit: int
       :question_list:


    For a POGIL or groupwork page to allow one partner to submit answers
    to all questions on the page for everyone in the group.
    :limit: sets the maximum group size.
    Question list is an anticipated extension that will allow the author
    to specify a list of questions rather than assuming all.
    The ``:optional:`` flag should almost always be set!!
    """

    required_arguments = 1
    optional_arguments = 1
    has_content = False
    option_spec = RunestoneDirective.option_spec.copy()
    option_spec.update(
        {
            "limit": directives.unchanged,
        }
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self):

        self.options["component_id"] = self.arguments[0].strip()

        if "limit" in self.options:
            self.options["size_limit"] = f"data-size_limit={self.options['limit']}"
        else:
            self.options["size_limit"] = f"data-size_limit=4"

        res = TEMPLATE.format(**self.options)

        return [nodes.raw(self.block_text, res, format="html")]



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = "isaiahmayerchak"

import os.path
import base64
import pathlib
from docutils import nodes
from docutils.parsers.rst import directives
from sqlalchemy import Table
from runestone.server.componentdb import get_engine_meta
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
)
import pdb


def setup(app):
    app.add_directive("datafile", DataFile)

    app.add_node(DataFileNode, html=(visit_df_html, depart_df_html))

    app.connect("doctree-resolved", process_datafile_nodes)
    app.connect("env-purge-doc", purge_datafiles)


TEMPLATE = """
<div class="runestone datafile %(optclass)s">
<div class="datafile_caption">Data file: <code>%(divid)s</code></div>
<pre data-component="datafile" id=%(divid)s %(hidden)s data-edit="%(edit)s" data-rows="%(rows)s" data-cols="%(cols)s">
%(filecontent)s</pre></div>
"""

IMG_TEMPLATE = """
<div class="runestone datafile">
<div class="datafile_caption">Data file: <code>%(divid)s</code></div>
<img data-component="datafile" data-isimage=%(image)s id=%(divid)s %(hidden)s src='data:image/%(imtype)s;base64, %(filecontent)s'/></div>
"""


class DataFileNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.


def visit_df_html(self, node):
    if "image" in node["runestone_options"]:
        res = IMG_TEMPLATE
    else:
        res = TEMPLATE
    res = res % node["runestone_options"]

    res = res.replace(
        "u'", "'"
    )  # hack:  there must be a better way to include the list and avoid unicode strings
    self.body.append(res)


def depart_df_html(self, node):
    """This is called at the start of processing an datafile node.  If datafile had recursive nodes
    etc and did not want to do all of the processing in visit_ac_html any finishing touches could be
    added here.
    """
    pass


def process_datafile_nodes(app, env, docname):
    pass


def purge_datafiles(app, env, docname):
    pass


class DataFile(RunestoneIdDirective):
    """
    .. datafile:: identifier
       :edit: Option that makes the datafile editable
       :cols: If editable, number of columns--default is 20
       :rows: If editable, number of rows--default is 40
       :hide: Flag that sets a non-editable datafile to be hidden
       :image: Flag that says this file is an image implies :fromfile: will be used
       :fromfile: path to file that contains the data
    """

    required_arguments = 1
    optional_arguments = 0
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "hide": directives.flag,
            "edit": directives.flag,
            "rows": directives.positive_int,
            "cols": directives.positive_int,
            "image": directives.flag,
            "fromfile": directives.unchanged,
        }
    )

    def run(self):
        """
        process the multiplechoice directive and generate html for output.
        :param self:
        :return:
        .. datafile:: identifier
            :edit: Option that makes the datafile editable
            :cols: If editable, number of columns--default is 20
            :rows: If editable, number of rows--default is 40
            :hide: Flag that sets a non-editable datafile to be hidden
            :image:
            :fromfile: path to file that contains the data
        """
        super(DataFile, self).run()
        env = self.state.document.settings.env

        if not hasattr(env, "datafilecounter"):
            env.datafilecounter = 0
        env.datafilecounter += 1
        import os

        if "fromfile" in self.options:
            ffpath = os.path.dirname(self.srcpath)
            filename = os.path.join(env.srcdir, ffpath, self.options["fromfile"])
            if "image" in self.options:
                self.options["imtype"] = pathlib.Path(filename).suffix[1:]
                with open(filename, "rb") as f:
                    self.content = base64.b64encode(f.read()).decode("utf8")
            else:
                with open(filename, "rb") as f:
                    self.content = [x[:-1].decode("utf8") for x in f.readlines()]
        if "cols" in self.options:
            self.options["cols"] = self.options["cols"]
        else:
            if "image" not in self.options:
                self.options["cols"] = min(65, max([len(x) for x in self.content]))
        if "rows" in self.options:
            self.options["rows"] = self.options["rows"]
        else:
            self.options["rows"] = 20

        if "image" in self.options:
            source = self.content
            self.options["image"] = "true"
        elif self.content:
            source = "\n".join(self.content) + "\n"
        else:
            source = "\n"
        self.options["filecontent"] = source

        if "hide" in self.options:
            self.options["hidden"] = "data-hidden"
        else:
            self.options["hidden"] = ""

        if "edit" in self.options:
            self.options["edit"] = "true"
        else:
            self.options["edit"] = "false"

        engine, meta, sess = get_engine_meta()
        if engine:
            Source_code = Table(
                "source_code", meta, autoload=True, autoload_with=engine
            )
            course_name = env.config.html_context["course_id"]
            divid = self.options["divid"]

            engine.execute(
                Source_code.delete()
                .where(Source_code.c.acid == divid)
                .where(Source_code.c.course_id == course_name)
            )
            engine.execute(
                Source_code.insert().values(
                    acid=divid, course_id=course_name, main_code=source
                )
            )
        else:
            print(
                "Unable to save to source_code table in datafile__init__.py. Possible problems:"
            )
            print("  1. dburl or course_id are not set in conf.py for your book")
            print("  2. unable to connect to the database using dburl")
            print()
            print(
                "This should only affect the grading interface. Everything else should be fine."
            )

        data_file_node = DataFileNode()
        data_file_node["runestone_options"] = self.options
        (
            data_file_node["source"],
            data_file_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)
        return [data_file_node]



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = "isaiahmayerchak"

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    maybeAddToAssignment,
)
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
)
from runestone.common.xmlcommon import write_substitute


def setup(app):
    app.add_directive("dragndrop", DragNDrop)

    app.add_node(
        DragNDropNode,
        html=(visit_dnd_html, depart_dnd_html),
        xml=(visit_dnd_xml, depart_dnd_xml),
    )

    app.add_config_value("dragndrop_div_class", "runestone", "html")


TEMPLATE_START = """
<div class="%(divclass)s  %(optclass)s">
<ul data-component="dragndrop" data-question_label="%(question_label)s" id="%(divid)s" %(randomize)s %(optional)s style="visibility: hidden;">
    <span data-subcomponent="question">%(qnumber)s: %(question)s</span>
	%(feedback)s
"""

TEMPLATE_OPTION = """
    <li data-subcomponent="draggable" id="%(divid)s_drag%(dnd_label)s">%(dragText)s</li>
    <li data-subcomponent="dropzone" for="%(divid)s_drag%(dnd_label)s">%(dropText)s</li>
"""
TEMPLATE_END = """</ul></div>"""

XML_START = """
<exercise label="{divid}">
    <statement><p>{question}</p></statement>
    {feedback}
"""


class DragNDropNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.


def visit_dnd_xml(self, node):
    res = XML_START

    if "feedback" in node["runestone_options"]:
        node["runestone_options"]["feedback"] = (
            "<feedback><p>" + node["runestone_options"]["feedback"] + "</p></feedback>"
        )
    else:
        node["runestone_options"]["feedback"] = ""

    res = res.format(**node["runestone_options"])
    self.output.append(res)


def depart_dnd_xml(self, node):
    self.output.append("<matches>")
    okeys = list(node["runestone_options"].keys())
    okeys.sort()
    for k in okeys:
        if "match" in k:
            x, ix = k.split("_")
            dragE, dropE = node["runestone_options"][k].split("|||")
            self.output.append(f'<match order="{ix}">')
            self.output.append(f"<premise>{dragE}</premise>")
            self.output.append(f"<response>{dropE}</response>")
            self.output.append("</match>")
    self.output.append("</matches></exercise>")


def visit_dnd_html(self, node):

    node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])
    self.body.append(node["delimiter"])

    res = visit_dnd_common(self, node)

    self.body.append(res)


def depart_dnd_html(self, node):
    res = depart_dnd_common(self, node)
    self.body.append(res)

    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        "".join(self.body[self.body.index(node["delimiter"]) + 1 :]),
    )

    self.body.remove(node["delimiter"])


def visit_dnd_common(self, node):
    res = TEMPLATE_START

    if "feedback" in node["runestone_options"]:
        node["runestone_options"]["feedback"] = (
            "<span data-subcomponent=feedback>"
            + node["runestone_options"]["feedback"]
            + "</span>"
        )
    else:
        node["runestone_options"]["feedback"] = ""

    # Add runestone_options : random : data-random to node
    if "randomize" in node["runestone_options"]:
        node["runestone_options"]["randomize"] = "data-random"
    else:
        node["runestone_options"]["randomize"] = ""
    

    res = res % node["runestone_options"]
    return res


def depart_dnd_common(self, node):
    res = ""
    # Add all of the possible answers
    okeys = list(node["runestone_options"].keys())
    okeys.sort()
    for k in okeys:
        if "match" in k:
            x, label = k.split("_")
            node["runestone_options"]["dnd_label"] = label
            dragE, dropE = node["runestone_options"][k].split("|||")
            node["runestone_options"]["dragText"] = dragE
            node["runestone_options"]["dropText"] = dropE
            res += node["template_option"] % node["runestone_options"]
    res += node["template_end"] % node["runestone_options"]

    return res


class DragNDrop(RunestoneIdDirective):
    """
    .. dragndrop:: identifier
        :feedback: Feedback that is displayed if things are incorrectly matched--is optional
        :match_1: Draggable element text|||Dropzone to be matched with text
        :match_2: Drag to Answer B|||Answer B
        :match_3: Draggable text|||Text of dropzone
        etc. (up to 20 matches)

        The question goes here.

    config values (conf.py):

    - dragndrop_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1
    optional_arguments = 1
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "feedback": directives.unchanged,
            "match_1": directives.unchanged,
            "match_2": directives.unchanged,
            "match_3": directives.unchanged,
            "match_4": directives.unchanged,
            "match_5": directives.unchanged,
            "match_6": directives.unchanged,
            "match_7": directives.unchanged,
            "match_8": directives.unchanged,
            "match_9": directives.unchanged,
            "match_10": directives.unchanged,
            "match_11": directives.unchanged,
            "match_12": directives.unchanged,
            "match_13": directives.unchanged,
            "match_14": directives.unchanged,
            "match_15": directives.unchanged,
            "match_16": directives.unchanged,
            "match_17": directives.unchanged,
            "match_18": directives.unchanged,
            "match_19": directives.unchanged,
            "match_20": directives.unchanged,
            "randomize": directives.flag,
        }
    )

    def run(self):
        """
        process the multiplechoice directive and generate html for output.
        :param self:
        :return:
        .. dragndrop:: identifier
            :feedback: Feedback that is displayed if things are incorrectly matched--is optional
            :match_1: Draggable element text|||Dropzone to be matched with text
            :match_2: Drag to Answer B|||Answer B
            :match_3: Draggable text|||Text of dropzone
            ...etc...(up to 20 matches)

            The question goes here.
        """
        super(DragNDrop, self).run()
        addQuestionToDB(self)

        if self.content:
            source = "\n".join(self.content)
        else:
            source = "\n"

        self.options["question"] = source
        env = self.state.document.settings.env
        self.options["divclass"] = env.config.dragndrop_div_class

        dndNode = DragNDropNode()
        dndNode["runestone_options"] = self.options
        dndNode["source"], dndNode["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        dndNode["template_start"] = TEMPLATE_START
        dndNode["template_option"] = TEMPLATE_OPTION
        dndNode["template_end"] = TEMPLATE_END

        maybeAddToAssignment(self)

        return [dndNode]



# *********************************
# |docname| - An indirect directive
# *********************************
#
# This directive lets you specify a question by random selection
# Given a list of question ids, it will randomly select one of those ids
# to present to the student.
# given a competency it will select a random question from all questions that
# test for that competency.

# Copyright (C) 2020  Runestone Academy LLC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = "bmiller"

# Imports from standard libarary
# ------------------------------

# Imports from third party libraries
# ----------------------------------
from docutils import nodes
from docutils.parsers.rst import directives

# local imports
# -------------
from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    maybeAddToAssignment,
)
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
)


TEMPLATE = """
<div class="runestone sqcontainer %(optclass)s">
<div data-component="selectquestion" id={component_id} {selector} {points} {proficiency} {min_difficulty} {max_difficulty} {autogradable} {not_seen_ever} {primary} {AB} {toggle_options} {toggle_labels} {limit_basecourse}>
    <p>{message}</p>
</div>
</div>
"""


def setup(app):
    app.add_directive("selectquestion", SelectQuestion)


class SelectQuestion(RunestoneIdDirective):
    """
    .. selectquestion:: uniqueid
       :fromid: [id [, id]+ ]
       :proficiency: randomly choose a question that tests a particular proficiency
       :primary: Only match on primary proficiencies
       :basecourse: restrict question choices to the current base course
       :alwaysrandom: choose a new random question every time if possible
       :points: number of points for this question
       :min_difficulty: minimum difficulty level
       :max_difficulty: maximum difficulty level
       :ab: experiment_name
       :toggle: allow student to choose which question to answer from the given list, with first question in fromid list being rendered first
       :togglelabels: relabel each question in the fromid list according to input here; blank will revert to a default label

       Difficulty is measured in one of two ways. For things like multiple choice and
       fill in the blank, we can use the % of students that get the answer correct on
       the first try.  For Code that is a little unreasonable, so we use the number
       of
    """

    required_arguments = 1
    optional_arguments = 1
    has_content = False
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "fromid": directives.unchanged,
            "proficiency": directives.unchanged,
            "basecourse": directives.flag,
            "min_difficulty": directives.unchanged,
            "max_difficulty": directives.unchanged,
            "autogradable": directives.flag,
            "not_seen_ever": directives.flag,
            "primary": directives.flag,
            "ab": directives.unchanged,
            "toggle": directives.unchanged,
            "togglelabels": directives.unchanged,
        }
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self):

        super(SelectQuestion, self).run()
        addQuestionToDB(self)
        env = self.state.document.settings.env
        is_dynamic = env.config.html_context.get("dynamic_pages", False)
        is_preview = env.config.html_context.get("course_id", None) == "preview"

        if not (bool("fromid" in self.options) ^ bool("proficiency" in self.options)):
            raise self.severe(
                "You must specify either fromid or proficiency but not both"
            )

        if is_dynamic or is_preview:
            self.options["message"] = "Loading a dynamic question ..."
        else:
            self.options["message"] = (
                "The selectquestion directive only works with Runestone Services"
            )

        if "fromid" in self.options:
            self.question_bank_choices = self.options["fromid"]
            self.options["selector"] = (
                f"data-questionlist='{self.question_bank_choices}'"
            )
            self.options[
                "message"
            ] += f"<br/>Selecting from: {self.question_bank_choices}"
        else:
            self.options["selector"] = ""

            # todo: validate that question(s) are in the database

        self.options["component_id"] = self.arguments[0].strip()

        if "proficiency" in self.options:
            self.options["proficiency"] = (
                f"""data-proficiency='{self.options["proficiency"]}'"""
            )
        else:
            self.options["proficiency"] = ""

        if "points" in self.options:
            self.int_points = int(self.options["points"])
            self.options["points"] = f"data-points={self.options['points']}"
        else:
            self.int_points = 1
            self.options["points"] = ""

        if "min_difficulty" in self.options:
            self.options["min_difficulty"] = (
                f"data-minDifficulty={self.options['min_difficulty']}"
            )
        else:
            self.options["min_difficulty"] = ""

        if "max_difficulty" in self.options:
            self.options["max_difficulty"] = (
                f"data-maxDifficulty={self.options['max_difficulty']}"
            )
        else:
            self.options["max_difficulty"] = ""

        if "autogradable" in self.options:
            self.options["autogradable"] = "data-autogradable=true"
        else:
            self.options["autogradable"] = ""

        if "not_seen_ever" in self.options:
            self.options["not_seen_ever"] = "data-not_seen_ever=true"
        else:
            self.options["not_seen_ever"] = ""

        if "primary" in self.options:
            self.options["primary"] = "data-primary=true"
        else:
            self.options["primary"] = ""

        if "ab" in self.options:
            if len(self.question_bank_choices.split(",")) != 2:
                raise self.severe("AB questions must have 2 options for :fromid:")
            self.options["AB"] = f"data-ab={self.options['ab']}"
        else:
            self.options["AB"] = ""

        if "basecourse" in self.options:
            self.options["limit_basecourse"] = "data-limit-basecourse=true"
        else:
            self.options["limit_basecourse"] = ""

        if ("toggle" in self.options) or ("togglelabels" in self.options):
            self.options["toggle_options"] = 'data-toggleoptions="toggle"'
            self.options["toggle_labels"] = 'data-togglelabels="togglelabels:"'
            if "toggle" in self.options:
                self.options["toggle_options"] = (
                    'data-toggleoptions="toggle, ' + f"{self.options['toggle']}" + '"'
                )
            if "togglelabels" in self.options:
                self.options["toggle_labels"] = (
                    'data-togglelabels="togglelabels: '
                    + f"{self.options['togglelabels']}"
                    + '"'
                )
        else:
            self.options["toggle_options"] = ""
            self.options["toggle_labels"] = ""

        maybeAddToAssignment(self)

        res = TEMPLATE.format(**self.options)
        addHTMLToDB(self.options["divid"], self.basecourse, res)

        return [nodes.raw(self.block_text, res, format="html")]



# Copyright (C) 2016  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "bmiller"


from collections import defaultdict
import binascii
import os
import re

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst import Directive
from docutils.utils import get_source_line
from docutils.statemachine import ViewList
from sphinx import application
from sphinx.errors import ExtensionError
from sphinx.util import logging


UNNUMBERED_DIRECTIVES = [
    #    "activecode",
    "reveal",
    #    "video",
    #    "youtube",
    #    "vimeo",
    #   "codelens",
    #   "showeval",
    #   "poll",
    "tabbed",
    "tab",
    "timed",
    "disqus",
]

logger = logging.getLogger(__name__)


# Provide a class which all Runestone nodes will inherit from.
class RunestoneNode(nodes.Node):
    pass


# Provide a class which all Runestone ID nodes will inherit from.
class RunestoneIdNode(RunestoneNode):
    pass


# Notes
# env = self.state.document.settings.env
# env.config.html_context['course_id']
# if not hasattr(env, 'activecodecounter'):
#    env.activecodecounter = 0
# env.activecodecounter += 1
# similar trick for assessments using getNumber()


# Easily create and initialize an object with named methods from the object's constructor. Taken from http://stackoverflow.com/a/3652937. While a collections.namedtuple could probably be forced to do this, the following approach seems fairly straightforward.
class Struct:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    # Provide a nice printable representation.
    def __repr__(self):
        args = ["{}={}".format(k, repr(v)) for (k, v) in vars(self).items()]
        return "Struct({})".format(", ".join(args))


# Get a data structure which holds Runestone data from the environment. Create one if it doesn't exist.
def _get_runestone_data(
    # The Sphinx environment which (possibly) contains Runestone data.
    env,
):

    # Create the Runestone data structure if it doesn't yet exist.
    if not hasattr(env, "runestone_data"):
        # .. _Runestone data:
        env.runestone_data = Struct(
            # Maps from a question's ID to the page containing it:
            #
            # .. code:: Python
            #   :number-lines:
            #
            #   .id_to_page = {
            #       # The ID maps to a Struct(), which contains:
            #       str(id): Struct()
            #           # The path of the file which contains this ID.
            #           .srcpath = str()
            #           # The line in this file which contains this ID.
            #           .line = int()
            #   }
            id_to_page=dict(),
            # Provide the inverse map: from a page to a set of IDs it contains.
            #
            # .. code:: Python
            #   :number-lines:
            #
            #   .page_to_id: = {
            #       # The path to a source file maps to a set:
            #       str(srcpath) : set(
            #           # Which consists of a strings, each of which is an ID on that page.
            #           str(id_), ...
            #       )
            #   }
            page_to_id=defaultdict(set),
        )
    return env.runestone_data


# When a source file is modified, clear all accumulated Runestone data from it.
def _purge_runestone_data(app, env, docname):
    runestone_data = _get_runestone_data(env)

    for id_ in runestone_data.page_to_id[docname]:
        runestone_data.id_to_page.pop(id_)
    runestone_data.page_to_id.pop(docname)

    # Reset the problem numbering for each page.
    env.assesscounter = 0


# Give the name of a file in the ``html_static_path``, append a version number after it based on its 32-bit CRC.
def _add_autoversion(
    self,
    # The file to find in the ``html_static_path``
    filename,
):

    # TODO: Cache the path and CRC to speed computing it.
    #
    # Search for this file by looking through all HTML static paths. The HTML builder hasn't been initilized, so the ``html_static_path`` doesn't show up yet as ``config.html_static_path``.
    # Some uses of this don't discriminate files versus urls. So if it starts with http just return it as we are not going to calculate a crc for something not on our server.
    if filename[:4] == "http":
        return filename
    for path in self.config._raw_config["html_static_path"]:
        full_path = path
        # Transform a relative path into an absolute path if necessary.
        if not os.path.isabs(full_path):
            full_path = os.path.join(self.confdir, full_path)
        # Distinguish between files and directories.
        if os.path.isdir(path):
            # Append the path to a directory.
            full_path = os.path.join(path, filename)
        # If it's not a directory, assume it's a file. See if these match.
        else:
            if os.path.normpath(path).endswith(os.path.normpath(filename)):
                full_path = path
            else:
                continue

        # See if we can open it.
        try:
            f = open(os.path.join(path, filename), "rb")
        except IOError:
            continue
        else:
            # Read the file and compute a CRC.
            with f:
                crc_str = "{:02X}".format(binascii.crc32(f.read()))

            # Append the CRC to the JavaScript reference.
            return "{}?v={}".format(filename, crc_str)

    # No match was found.
    raise ExtensionError("Unable to find {} in html_static_path.".format(filename))


# Convenience method to call ``add_js_file`` using ``add_autoversion``.
def _add_autoversioned_javascript(self, filename, **kwargs):
    return self.add_js_file(self.add_autoversion(filename), **kwargs)


# Convenience method for calling ``add_css_file`` using ``add_autoversion``.
def _add_autoversioned_stylesheet(self, filename, *args, **kwargs):
    return self.add_css_file(self.add_autoversion(filename), *args, **kwargs)


# Add these methods to the Sphinx application object.
application.Sphinx.add_autoversion = _add_autoversion
application.Sphinx.add_autoversioned_javascript = _add_autoversioned_javascript
application.Sphinx.add_autoversioned_stylesheet = _add_autoversioned_stylesheet


def setup(app):
    # Avoid a circular import. Ick.
    from .question_number import _insert_qnum

    # See http://www.sphinx-doc.org/en/stable/extdev/appapi.html#event-env-purge-doc.
    app.connect("env-purge-doc", _purge_runestone_data)
    app.connect("doctree-resolved", _insert_qnum)
    app.add_role("skipreading", SkipReading)
    # See http://www.sphinx-doc.org/en/stable/extdev/appapi.html#sphinx.application.Sphinx.add_config_value.
    app.add_config_value("runestone_server_side_grading", False, "env")
    app.add_config_value("generate_component_labels", True, "env")


# Roughly what an XML ID allows, per the spec for an `XML Name
# <https://www.w3.org/TR/REC-xml/#NT-Name>`_. However this is a bit more restrictive:
#
# - This disallows characters that need escaping for CSS to avoid problems there, such as ``.``.
# - There are Unicode code points allowed in a NCName that are not matched by \w.
xml_id_regex = re.compile(r'^(?!\d)[\w_][\w-]*$', re.UNICODE)



# A base class for all Runestone directives.
class RunestoneDirective(Directive):
    option_spec = {
        "author": directives.unchanged,
        "tags": directives.unchanged,
        "class": directives.unchanged,
        "difficulty": directives.unchanged,
        "autograde": directives.unchanged,
        "practice": directives.unchanged,
        "topics": directives.unchanged,
        "optional": directives.flag,
        "prim_comp": directives.unchanged,
        "supp_comp": directives.unchanged,
        "from_source": directives.unchanged,
        "basecourse": directives.unchanged,
        "chapter": directives.unchanged,
        "subchapter": directives.unchanged,
        "points": directives.positive_int,
        "pct_on_first": directives.unchanged,
        "mean_clicks_to_correct": directives.unchanged,
        "total_students_attempting": directives.unchanged,
        "num_students_correct": directives.unchanged,
    }

    def __init__(self, *args, **kwargs):
        super(RunestoneDirective, self).__init__(*args, **kwargs)
        env = self.state.document.settings.env
        self.srcpath = env.docname
        # Rather tha use ``os.sep`` to split ``self.srcpath``, use ``'/'``, because Sphinx internally stores filesnames using this separator, even on Windows.

        split_docname = self.srcpath.split("/")
        if len(split_docname) < 2:
            # TODO: Warn about this? Something like ``self.state.document.settings.env``?
            split_docname.append("")
        if "subchapter" not in self.options:
            self.subchapter = split_docname[-1]
            self.options["subchapter"] = self.subchapter
        else:
            self.subchapter = self.options["subchapter"]
        if "chapter" not in self.options:
            self.chapter = split_docname[-2]
            self.options["chapter"] = self.chapter
        else:
            self.chapter = self.options["chapter"]

        if "basecourse" not in self.options:
            self.basecourse = self.state.document.settings.env.config.html_context.get(
                "basecourse", "unknown"
            )
            self.options["basecourse"] = self.basecourse
        self.options["optional"] = (
            "data-optional=true" if "optional" in self.options else ""
        )
        if "points" in self.options:
            self.int_points = int(self.options["points"])
        else:
            self.int_points = 1

        self.options["optclass"] = self.options.get("class", "")

        self.explain_text = []

    # Check for a valid XML id. This is more restrictive then checking for a `valid HTML5 divid <https://html.spec.whatwg.org/multipage/dom.html#the-id-attribute>`_, so we don't bother with a separate HTML ID check.
    def validate_divid(self, divid):
        if not xml_id_regex.match(divid):
            logger.error(
                f"Invalid divid '{divid}'.",
                location=self.state_machine.get_source_and_line(self.lineno),
            )


# This is a base class for all Runestone directives which require a divid as their first parameter.
class RunestoneIdDirective(RunestoneDirective):
    def getNumber(self):
        env = self.state.document.settings.env

        if (
            self.name in UNNUMBERED_DIRECTIVES
            or env.config.generate_component_labels is False
        ):
            return ""

        env.assesscounter += 1
        # print(f" assesscounter = {env.assesscounter}")
        res = "Q-%d"

        if hasattr(env, "assessprefix"):
            res = env.assessprefix + "%d"

        res = res % env.assesscounter

        if hasattr(env, "assesssuffix"):
            res += env.assesssuffix

        return res

    def updateContent(self):
        if self.content:
            # If the first line is a directive, then put the numbering on a separate line.
            if self.content[0][:2] == "..":
                # Per the `docs <http://www.sphinx-doc.org/en/stable/extdev/markupapi.html#viewlists>`_, ``self.content`` is a ``ViewList``, so we can't just insert strings. Instead, create a viewlist the combine the two.
                self.content = (
                    # Use the source file from the existing ViewList when creating a new one.
                    ViewList(
                        [self.options["qnumber"] + ":", ""], self.content.source(0)
                    )
                    + self.content
                )
            else:
                if self.options["qnumber"]:
                    self.content[0] = self.options["qnumber"] + ": " + self.content[0]

    def get_explain_text(self):
        try:
            idx = self.content.index("")  # find the first blank line in the contents
        except ValueError:
            idx = len(self.content)
        return self.content[:idx]

    def run(self):
        # Make sure the runestone directive at least requires an ID.
        assert self.required_arguments >= 1
        if "divid" not in self.options:
            divid = self.options["divid"] = self.arguments[0]
        else:
            divid = self.options["divid"]
        self.validate_divid(divid)

        self.options["qnumber"] = self.getNumber()
        # print(f"{id_} is number {self.options['qnumber']}")

        # Get references to `runestone data`_.
        env = self.state.document.settings.env
        runestone_data = _get_runestone_data(env)
        id_to_page = runestone_data.id_to_page
        page_to_id = runestone_data.page_to_id
        # See if this ID already exists.
        if divid in id_to_page:
            page = id_to_page[divid]
            # If it's not simply an update to an existing ID, complain.
            if page.docname != env.docname or page.lineno != self.lineno:
                raise self.error(
                    "Duplicate ID -- see {}, line {}".format(page.docname, page.lineno)
                )
            # Make sure our data structure is consistent.
            assert divid in page_to_id[page.docname]
        else:
            # Add a new entry.
            id_to_page[divid] = Struct(docname=env.docname, lineno=self.lineno)
            page_to_id[env.docname].add(divid)

        self.in_exam = getattr(env, "in_timed", "")


# returns True when called first time with particular parameters' values
def first_time(app, *keys):
    key = "$".join(keys)
    if not hasattr(app, "runestone_flags"):
        app.runestone_flags = set()
    if not key in app.runestone_flags:
        app.runestone_flags.add(key)
        return True
    return False


# Some nodes have a line number of None. Look through their children to find the node's line number.
def get_node_line(node):
    # This returns source, line. Pick just the line.
    return get_source_line(node)[1]


def SkipReading(
    # _`roleName`: the local name of the interpreted role, the role name actually used in the document.
    roleName,
    # _`rawtext` is a string containing the enitre interpreted text input, including the role and markup. Return it as a problematic node linked to a system message if a problem is encountered.
    rawtext,
    # The interpreted _`text` content.
    text,
    # The line number (_`lineno`) where the interpreted text begins.
    lineno,
    # _`inliner` is the docutils.parsers.rst.states.Inliner object that called this function. It contains the several attributes useful for error reporting and document tree access.
    inliner,
    # A dictionary of directive _`options` for customization (from the "role" directive), to be interpreted by this function. Used for additional attributes for the generated elements and other functionality.
    options={},
    # A list of strings, the directive _`content` for customization (from the "role" directive). To be interpreted by the role function.
    content=[],
):

    docname = inliner.document.settings.env.docname

    if not hasattr(inliner.document.settings.env, "skipreading"):
        inliner.document.settings.env.skipreading = set()

    # print("ADDING {} to skipreading".format(docname))
    inliner.document.settings.env.skipreading.add(docname)

    return ([], [])


# Return the Sphinx environment given a node object.
def env_from_node(node):
    # Ascend the node tree until we find a node with a ``document``.
    node_with_document = node
    while not node_with_document.document:
        node_with_document = node_with_document.parent
    return node_with_document.document.settings.env
